import dataclasses

from rich.console import Console

from apolo_flow.parser import ConfigDir


@dataclasses.dataclass(frozen=True)
class Root:
    config_dir: ConfigDir
    console: Console
    verbosity: int
    show_traceback: bool



# expression parser/evaluator
# ${{ <expression> }}
import dataclasses

import abc
import asyncio
import collections
import collections.abc
import datetime
import enum
import hashlib
import inspect
import json
import operator
import re
import shlex
from abc import ABC
from apolo_sdk import Client, JobDescription, JobRestartPolicy, JobStatus
from ast import literal_eval
from collections.abc import Sized
from contextlib import asynccontextmanager
from funcparserlib.parser import (
    Parser,
    finished,
    forward_decl,
    many,
    maybe,
    oneplus,
    skip,
    some,
)
from typing import (
    Any,
    AsyncContextManager,
    AsyncIterator,
    Awaitable,
    Callable,
    ClassVar,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
)
from typing_extensions import Final, Protocol, runtime_checkable
from yarl import URL

from .tokenizer import Pos, Token, tokenize
from .types import AlwaysT, LocalPath, RemotePath, TaskStatus
from .utils import run_subproc


LiteralT = Union[None, bool, int, float, str]

TypeT = Union[
    LiteralT,
    "ContainerT",
    "MappingT",
    "SequenceT",
    LocalPath,
    RemotePath,
    URL,
    AlwaysT,
]

_T = TypeVar("_T", bound=TypeT)
_IT = TypeVar("_IT", bound=TypeT)


@runtime_checkable
class ContainerT(Protocol):
    def __getattr__(self, attr: str) -> TypeT: ...


@runtime_checkable
class MappingT(Protocol):
    def __getitem__(self, key: LiteralT) -> TypeT: ...

    def __iter__(self) -> Iterator[LiteralT]: ...


@runtime_checkable
class SequenceT(Protocol):
    def __getitem__(self, idx: LiteralT) -> TypeT: ...

    def __iter__(self) -> Iterator[TypeT]: ...


class RootABC(abc.ABC):
    @abc.abstractmethod
    def lookup(self, name: str) -> TypeT:
        pass

    @abc.abstractmethod
    def client(self) -> AsyncContextManager[Client]:
        pass

    @property
    @abc.abstractmethod
    def dry_run(self) -> bool: ...


class LocalScope(RootABC):
    def __init__(self, base: RootABC, scope: Mapping[str, TypeT]) -> None:
        self._base = base
        self._scope = scope

    @asynccontextmanager
    async def client(self) -> AsyncIterator[Client]:
        async with self._base.client() as client:
            yield client

    @property
    def dry_run(self) -> bool:
        return self._base.dry_run

    def lookup(self, name: str) -> TypeT:
        if name in self._scope:
            return self._scope[name]
        return self._base.lookup(name)


class EvalError(Exception):
    def __init__(self, msg: str, start: Pos, end: Pos) -> None:
        super().__init__(msg)
        self.start = start
        self.end = end

    @staticmethod
    def format_pos(pos: Pos) -> str:
        # For humans, line and columns are enumerated from 1, so we should add 1 here.
        return f'"{pos.filename}", line {pos.line + 1}, column {pos.col + 1}'

    def __str__(self) -> str:
        return str(self.args[0]) + f"\n  in {self.format_pos(self.start)}"


class MultiError(Exception):
    def __init__(self, errors: Sequence[Exception]):
        self.errors = errors

    def __str__(self) -> str:
        return "\n".join(str(error) for error in self.errors)


def parse_literal(arg: str, err_msg: str) -> LiteralT:
    try:
        return cast(LiteralT, literal_eval(arg))
    except (ValueError, SyntaxError):
        raise ValueError(f"'{arg}' is not " + err_msg)


@dataclasses.dataclass(frozen=True)
class FuncDef:
    name: str
    sig: inspect.Signature
    call: Callable[..., Awaitable[TypeT]]


@dataclasses.dataclass
class CallCtx:
    start: Pos
    end: Pos
    root: RootABC


def _build_signatures(**kwargs: Callable[..., Awaitable[TypeT]]) -> Dict[str, FuncDef]:
    return {k: FuncDef(k, inspect.signature(v), v) for k, v in kwargs.items()}


async def nothing(ctx: CallCtx) -> None:
    # A test function that accepts none args.
    # Later we can replace it with something really more usefuld, e.g. succeded()
    return None


async def alen(ctx: CallCtx, arg: TypeT) -> int:
    # Async version of len(), async is required for the sake of uniformness.
    if not isinstance(arg, Sized):
        raise TypeError(f"len() requires a str, sequence or mapping, got {arg!r}")
    return len(arg)


async def aint(ctx: CallCtx, arg: TypeT) -> int:
    # Async version of int(), async is required for the sake of uniformness.
    if not isinstance(arg, (int, str, float, bool)):
        raise TypeError(f"int() requires a str or a number, got {arg!r}")
    return int(arg)


async def akeys(ctx: CallCtx, arg: TypeT) -> TypeT:
    # Async version of list(), async is required for the sake of uniformness.
    if not isinstance(arg, Mapping):
        raise TypeError(f"keys() requires a mapping, got {arg!r}")
    return list(arg)  # type: ignore  # List[...] is implicitly converted to SequenceT


async def values(ctx: CallCtx, arg: TypeT) -> TypeT:
    # Trampoline for mapping.values, async is required for the sake of uniformness.
    if not isinstance(arg, Mapping):
        raise TypeError(f"values() requires a mapping, got {arg!r}")
    return list(arg.values())  # type: ignore  # implicitly converted to SequenceT


async def arange(
    ctx: CallCtx, arg: TypeT, arg2: Optional[TypeT] = None, arg3: Optional[TypeT] = None
) -> SequenceT:
    if not isinstance(arg, int):
        raise TypeError(f"range() first argument should be an int, got {arg!r}")
    if arg2:
        if not isinstance(arg2, int):
            raise TypeError(f"range() second argument should be an int, got {arg2!r}")
        if arg3:
            if not isinstance(arg3, int):
                raise TypeError(
                    f"range() third argument should be an int, got {arg3!r}"
                )
            return list(range(arg, arg2, arg3))  # type: ignore
        return list(range(arg, arg2))  # type: ignore
    return list(range(arg))  # type: ignore


async def fmt(ctx: CallCtx, spec: str, *args: TypeT) -> str:
    # We need a trampoline since expression syntax doesn't support classes and named
    # argumens
    return spec.format(*args)


class JSONEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        if not isinstance(obj, type) and dataclasses.is_dataclass(obj):
            return dataclasses.asdict(obj)
        elif isinstance(obj, enum.Enum):
            return obj.value
        elif isinstance(obj, RemotePath):
            return str(obj)
        elif isinstance(obj, LocalPath):
            return str(obj)
        elif isinstance(obj, collections.abc.Set):
            return sorted(obj)
        elif isinstance(obj, AlwaysT):
            return str(obj)
        elif isinstance(obj, URL):
            return str(obj)
        elif isinstance(obj, datetime.datetime):
            return str(obj)
        # Let the base class default method raise the TypeError
        return super().default(obj)


async def to_json(ctx: CallCtx, arg: TypeT) -> str:
    return json.dumps(arg, cls=JSONEncoder)


async def from_json(ctx: CallCtx, arg: str) -> TypeT:
    return cast(TypeT, json.loads(arg))


async def hash_files(ctx: CallCtx, *patterns: str) -> str:
    flow = ctx.root.lookup("flow")
    # emulate attr lookup
    workspace: LocalPath = cast(
        LocalPath,
        await AttrGetter(ctx.start, ctx.end, name="workspace").eval(
            ctx.root, flow, start=ctx.start
        ),
    )
    return await hash_files_relative(ctx, workspace, *patterns)


async def hash_files_relative(ctx: CallCtx, root: LocalPath, *patterns: str) -> str:
    hasher = hashlib.new("sha256")
    buffer = bytearray(16 * 1024 * 1024)  # 16 MB
    view = memoryview(buffer)
    # emulate attr lookup
    for pattern in patterns:
        for fname in sorted(root.glob(pattern)):
            # On Windows the Python 3.6 glob() returns lower-cased filenames,
            # resolve() restores the case.
            relative_fname = fname.resolve().relative_to(root.resolve()).as_posix()
            hasher.update(relative_fname.encode("utf-8"))
            with fname.open("rb", buffering=0) as stream:
                read = stream.readinto(buffer)
                while read:
                    hasher.update(view[:read])
                    read = stream.readinto(buffer)
    return hasher.hexdigest()


async def upload(ctx: CallCtx, volume_ctx: ContainerT) -> ContainerT:
    from .context import VolumeCtx

    if not isinstance(volume_ctx, VolumeCtx):
        raise ValueError("upload() argument should be volume")
    mkdir_cmd = [
        "apolo",
        "mkdir",
        "--parents",
        str(volume_ctx.remote.parent),
    ]
    cp_cmd = [
        "apolo",
        "cp",
        "--recursive",
        "--update",
        "--no-target-directory",
        str(volume_ctx.full_local_path),
        str(volume_ctx.remote),
    ]
    if ctx.root.dry_run:
        print(" ".join(shlex.quote(arg) for arg in mkdir_cmd))
        print(" ".join(shlex.quote(arg) for arg in cp_cmd))
    else:
        await run_subproc(*mkdir_cmd)
        await run_subproc(*cp_cmd)
    return volume_ctx


async def alower(ctx: CallCtx, arg: TypeT) -> str:
    # Async version of lower(), async is required for the sake of uniformness.
    if not isinstance(arg, str):
        raise TypeError(f"lower() requires a str, got {arg!r}")
    return arg.lower()


async def aupper(ctx: CallCtx, arg: TypeT) -> str:
    # Async version of lower(), async is required for the sake of uniformness.
    if not isinstance(arg, str):
        raise TypeError(f"upper() requires a str, got {arg!r}")
    return arg.upper()


async def astr(ctx: CallCtx, arg: TypeT) -> str:
    # Async version of str(), async is required for the sake of uniformness.
    return str(arg)


async def replace(ctx: CallCtx, arg: TypeT, old: TypeT, new: TypeT) -> str:
    # We need a trampoline since expression syntax doesn't support classes
    if not isinstance(arg, str):
        raise TypeError(f"replace() first argument should be a str, got {arg!r}")
    if not isinstance(old, str):
        raise TypeError(f"replace() second argument should be a str, got {old!r}")
    if not isinstance(new, str):
        raise TypeError(f"replace() third argument should be a str, got {new!r}")
    return arg.replace(old, new)


async def join(
    ctx: CallCtx, sep: TypeT, array: TypeT, convert_to_str: TypeT = False
) -> str:
    # We need a trampoline since expression syntax doesn't support classes
    if not isinstance(sep, str):
        raise TypeError(f"join() first argument should be a str, got {sep!r}")
    if not isinstance(array, SequenceT):
        raise TypeError(
            f"replace() second argument should be a sequence, got {array!r}"
        )
    if not isinstance(convert_to_str, bool):
        raise TypeError(
            f"replace() third argument should be a bool, got {convert_to_str!r}"
        )
    str_array = []
    for idx, item in enumerate(array):
        if convert_to_str:
            item = str(item)
        if not isinstance(item, str):
            raise TypeError(f"join() array item {idx} should be a str, got {item!r}")
        str_array.append(item)
    return sep.join(str_array)


async def parse_volume(ctx: CallCtx, arg: TypeT) -> ContainerT:
    from .context import VolumeCtx

    if not isinstance(arg, str):
        raise TypeError(f"parse_volume() requires a str, got {arg!r}")
    async with ctx.root.client() as client:
        volume = client.parse.volume(arg)
    return VolumeCtx(  # type: ignore[return-value]
        id="<volume>",
        remote=volume.storage_uri,
        mount=RemotePath(volume.container_path),
        read_only=volume.read_only,
        local=None,
        full_local_path=None,
    )


def _check_has_needs(ctx: CallCtx, *, func_name: str) -> None:
    try:
        ctx.root.lookup("needs")
    except LookupError:
        raise ValueError(f"{func_name}() is only available inside a task definition")


def _get_needs_statuses(root: RootABC) -> Dict[str, TaskStatus]:
    needs = root.lookup("needs")
    assert isinstance(needs, MappingT)
    result: Dict[str, TaskStatus] = {}
    for dependency in needs:
        dep_ctx = needs[dependency]
        result[dependency] = dep_ctx.result  # type: ignore
    return result


async def always(ctx: CallCtx, *args: str) -> AlwaysT:
    _check_has_needs(ctx, func_name="always")
    return AlwaysT()


async def success(ctx: CallCtx, *args: str) -> bool:
    # When called without arguments, checks that all dependency task
    # succeeded. If arguments are passed, they should be strings that
    # name some task from `needs:` field. Function will return true
    # if all of those tasks succeded.
    _check_has_needs(ctx, func_name="success")
    needs_statuses = _get_needs_statuses(ctx.root)
    if not args:
        return all(status == TaskStatus.SUCCEEDED for status in needs_statuses.values())
    else:
        if not all(isinstance(arg, str) for arg in args):
            raise ValueError("success() function only accept string arguments")
        try:
            return all(needs_statuses[need] == TaskStatus.SUCCEEDED for need in args)
        except KeyError as e:
            raise ValueError(
                "success() function got argument, that"
                f'is not a defined as dependency: "{e.args[0]}"'
            )


async def failure(ctx: CallCtx, *args: str) -> bool:
    # When called without arguments, checks if any of dependency task
    # failed. If arguments are passed, they should be strings that
    # name some task from `needs:` field. Function will return true
    # if any of those tasks failure.
    _check_has_needs(ctx, func_name="failure")
    needs_statuses = _get_needs_statuses(ctx.root)
    if not args:
        return any(status == TaskStatus.FAILED for status in needs_statuses.values())
    else:
        if not all(isinstance(arg, str) for arg in args):
            raise ValueError("failure() function only accept string arguments")
        try:
            return any(needs_statuses[need] == TaskStatus.FAILED for need in args)
        except KeyError as e:
            raise ValueError(
                "failure() function got argument, that"
                f'is not a defined as dependency: "{e.args[0]}"'
            )


async def inspect_job(
    ctx: CallCtx, job_name: TypeT, suffix: Optional[TypeT] = None
) -> ContainerT:
    from .context import LiveContext, _id2tag

    if not isinstance(ctx.root, LiveContext):
        raise ValueError("inspect_job() is only available inside a job definition")
    if not isinstance(job_name, str):
        raise TypeError(
            f"inspect_job() job_name argument should be a str, got {job_name!r}"
        )
    if suffix is not None and not isinstance(suffix, str):
        raise TypeError(
            f"inspect_job() suffix argument should be a str, got {suffix!r}"
        )

    tags = ctx.root.tags | {f"job:{_id2tag(job_name)}"}
    if suffix is not None:
        tags |= {f"multi:{suffix}"}
    found_job: Optional[JobDescription] = None
    async with ctx.root.client() as client:
        async for job in client.jobs.list(
            tags=tags,
            reverse=True,
            limit=1,
            statuses=JobStatus.active_items(),
        ):
            found_job = job
    if found_job is None:
        raise ValueError(
            f"inspect_job() did not found running job with name {job_name}"
            + (f" and suffix {suffix}" if suffix else "")
        )
    return found_job  # type: ignore[return-value]


FUNCTIONS = _build_signatures(
    len=alen,
    int=aint,
    range=arange,
    nothing=nothing,
    fmt=fmt,
    str=astr,
    replace=replace,
    join=join,
    keys=akeys,
    values=values,
    to_json=to_json,
    from_json=from_json,
    success=success,
    failure=failure,
    hash_files=hash_files,
    hash_files_relative=hash_files_relative,
    always=always,
    upload=upload,
    lower=alower,
    upper=aupper,
    parse_volume=parse_volume,
    inspect_job=inspect_job,
)


@dataclasses.dataclass(frozen=True)
class Entity(abc.ABC):
    start: Pos
    end: Pos


class Item(Entity):
    @abc.abstractmethod
    async def eval(self, root: RootABC) -> TypeT:
        pass

    def child_items(self) -> Iterable["Item"]:
        return []


@dataclasses.dataclass(frozen=True)
class Literal(Item):
    val: LiteralT

    async def eval(self, root: RootABC) -> LiteralT:
        return self.val


def make_toktype_predicate(toktype: str) -> Callable[[Token], bool]:
    def _predicate(token: Token) -> bool:
        return token.type == toktype

    return _predicate


def literal(toktype: str) -> "Parser[Token, Literal]":
    def f(tok: Token) -> Literal:
        return Literal(tok.start, tok.end, literal_eval(tok.value))

    return some(make_toktype_predicate(toktype)) >> f


class Getter(Entity):
    # Aux class for Lookup item

    @abc.abstractmethod
    async def eval(self, root: RootABC, obj: TypeT, start: Pos) -> TypeT:
        pass


@dataclasses.dataclass(frozen=True)
class AttrGetter(Getter):
    name: str

    async def eval(self, root: RootABC, obj: TypeT, start: Pos) -> TypeT:
        if dataclasses.is_dataclass(obj):
            name = self.name
            try:
                return cast(TypeT, getattr(obj, name))
            except AttributeError:
                raise EvalError(f"No attribute {self.name}", start, self.end)
        elif isinstance(obj, MappingT):
            try:
                return obj[self.name]
            except KeyError:
                raise EvalError(f"No attribute {self.name}", start, self.end)
        else:
            raise EvalError(
                f"Is not an object with attributes accessible by a dot.",
                start,
                self.end,
            )


def lookup_attr(name: Token) -> Any:
    # Just in case, NAME token cannot start with _.
    assert not name.value.startswith("_")
    return AttrGetter(name.start, name.end, name.value)


@dataclasses.dataclass(frozen=True)
class ItemGetter(Getter):
    key: Item

    async def eval(self, root: RootABC, obj: TypeT, start: Pos) -> TypeT:
        assert isinstance(obj, (SequenceT, MappingT))
        key = cast(LiteralT, await self.key.eval(root))
        try:
            return obj[key]
        except LookupError:
            raise EvalError(f"No item {self.key}", start, self.end)


def lookup_item(key: Item) -> Any:
    return ItemGetter(key.start, key.end, key)


@dataclasses.dataclass(frozen=True)
class Lookup(Item):
    root: str
    trailer: Sequence[Getter]

    async def eval(self, root: RootABC) -> TypeT:
        ret = root.lookup(self.root)
        start = self.start
        for op in self.trailer:
            ret = await op.eval(root, ret, start)
        return ret

    def child_items(self) -> Iterable["Item"]:
        for getter in self.trailer:
            if isinstance(getter, ItemGetter):
                yield getter.key


def make_lookup(arg: Tuple[Token, List[Getter]]) -> Lookup:
    name, trailer = arg
    end = trailer[-1].end if trailer else name.end
    return Lookup(name.start, end, name.value, trailer)


def make_args(arg: Optional[Tuple[Item, List[Item]]]) -> List[Item]:
    if arg is None:
        return []
    first, tail = arg
    return [first] + tail[:]


@dataclasses.dataclass(frozen=True)
class Call(Item):
    func: FuncDef
    args: Sequence[Item]
    trailer: Sequence[Getter]

    async def eval(self, root: RootABC) -> TypeT:
        args = [await a.eval(root) for a in self.args]
        try:
            call_ctx = CallCtx(self.start, self.end, root)
            ret = await self.func.call(call_ctx, *args)
        except asyncio.CancelledError:
            raise
        except EvalError:
            raise
        except Exception as exc:
            raise EvalError(str(exc), self.start, self.end)
        start = self.start
        for op in self.trailer:
            ret = await op.eval(root, ret, start)
        return ret

    def child_items(self) -> Iterable["Item"]:
        yield from self.args
        for getter in self.trailer:
            if isinstance(getter, ItemGetter):
                yield getter.key


def make_call(arg: Tuple[Token, List[Item], Sequence[Getter]]) -> Call:
    funcname, args, trailer = arg
    try:
        spec = FUNCTIONS[funcname.value]
    except KeyError:
        raise LookupError(f"Unknown function {funcname.value}")
    end = funcname.end
    args_count = len(args)
    if args_count > 0:
        end = args[-1].end
    dummies = [None] * args_count
    spec.sig.bind(None, *dummies)
    if trailer:
        end = trailer[-1].end
    return Call(funcname.start, end, spec, args, trailer)


@dataclasses.dataclass(frozen=True)
class Text(Item):
    arg: str

    async def eval(self, root: RootABC) -> str:
        return self.arg


def make_text(arg: Token) -> Text:
    return Text(arg.start, arg.end, arg.value)


@dataclasses.dataclass(frozen=True)
class BinOp(Item):
    op: Callable[[TypeT, TypeT], TypeT]
    left: Item
    right: Item

    async def eval(self, root: RootABC) -> TypeT:
        left_val = await self.left.eval(root)
        right_val = await self.right.eval(root)
        return self.op(left_val, right_val)

    def child_items(self) -> Iterable["Item"]:
        return [self.left, self.right]


def or_(arg1: Any, arg2: Any) -> Any:
    # Emulate dicts concatination from Python 3.9
    d: Mapping[Any, Any]
    if isinstance(arg1, collections.abc.Mapping):
        d = dict(arg1)
        d.update(arg2)
        return d
    elif isinstance(arg2, collections.abc.Mapping):
        d = dict(arg2)
        d.update(arg1)
        return d
    else:
        return operator.or_(arg1, arg2)


def logical_and(arg1: Any, arg2: Any) -> Any:
    return arg1 and arg2


def logical_or(arg1: Any, arg2: Any) -> Any:
    return arg1 or arg2


BinOpTrailer = List[Tuple[Token, Item]]


def make_op_trailer(args: Optional[Tuple[Token, Item, BinOpTrailer]]) -> BinOpTrailer:
    if args is None:
        return []
    op_token, item, trailer = args
    return [(op_token, item), *trailer]


def make_bin_op_expr(args: Tuple[Item, BinOpTrailer]) -> Item:
    op_map = {
        "==": operator.eq,
        "!=": operator.ne,
        "or": logical_or,
        "and": logical_and,
        "<": operator.lt,
        "<=": operator.le,
        ">": operator.gt,
        ">=": operator.ge,
        "|": or_,
        "+": operator.add,
        "-": operator.sub,
        "*": operator.mul,
        "/": operator.truediv,
        "%": operator.mod,
    }

    item, trailer = args
    for op_token, right_item in trailer:
        item = BinOp(
            item.start,
            right_item.end,
            op=op_map[op_token.value],
            left=item,
            right=right_item,
        )
    return item


@dataclasses.dataclass(frozen=True)
class UnaryOp(Item):
    op: Callable[[TypeT], TypeT]
    operand: Item

    async def eval(self, root: RootABC) -> TypeT:
        operand_val = await self.operand.eval(root)
        return self.op(operand_val)

    def child_items(self) -> Iterable["Item"]:
        return [self.operand]


def _unary_plus(arg: Any) -> Any:
    return +arg


def _unary_minus(arg: Any) -> Any:
    return -arg


def make_unary_op_expr(args: Tuple[Token, Item]) -> UnaryOp:
    op_map = {
        "not": operator.not_,
        "+": _unary_plus,
        "-": _unary_minus,
    }
    op_token = args[0]
    return UnaryOp(
        args[0].start,
        args[1].end,
        op=op_map[op_token.value],
        operand=args[1],
    )


@dataclasses.dataclass(frozen=True)
class ListMaker(Item):
    items: Sequence[Item]

    async def eval(self, root: RootABC) -> SequenceT:
        return [await item.eval(root) for item in self.items]  # type: ignore

    def child_items(self) -> Iterable["Item"]:
        return self.items


def make_list(args: Tuple[Item, List[Item]]) -> ListMaker:
    lst = [args[0]] + args[1]
    return ListMaker(lst[0].start, lst[-1].end, lst)


def make_empty_list(args: Tuple[Token, Token]) -> ListMaker:
    return ListMaker(args[0].start, args[1].end, [])


@dataclasses.dataclass(frozen=True)
class ListCompMaker(Item):
    item_expr: Item
    var_name: str
    base_iter: Item
    if_expr: Optional[Item] = None

    async def eval(self, root: RootABC) -> SequenceT:
        res = []
        for value in await self.base_iter.eval(root):  # type: ignore
            subroot = LocalScope(root, {self.var_name: value})
            if self.if_expr and not await self.if_expr.eval(subroot):
                continue
            res.append(await self.item_expr.eval(subroot))
        return res  # type: ignore

    def child_items(self) -> Iterable["Item"]:
        return [self.item_expr, self.base_iter]


def make_list_comp(args: Tuple[Item, Token, Token, Token, Item]) -> ListCompMaker:
    item_expr, _, name, _, base_iter = args
    return ListCompMaker(
        item_expr.start,
        base_iter.end,
        item_expr,
        name.name,
        base_iter,
    )


def make_list_comp_with_if(
    args: Tuple[Item, Token, Token, Token, Item, Token, Item]
) -> ListCompMaker:
    item_expr, _, name, _, base_iter, _, if_expr = args
    return ListCompMaker(
        item_expr.start,
        base_iter.end,
        item_expr,
        name.name,
        base_iter,
        if_expr,
    )


@dataclasses.dataclass(frozen=True)
class DictMaker(Item):
    items: Sequence[Tuple[Item, Item]]

    async def eval(self, root: RootABC) -> MappingT:
        ret = {}
        for key, value in self.items:
            k = await key.eval(root)
            v = await value.eval(root)
            ret[k] = v
        return ret  # type: ignore

    def child_items(self) -> Iterable["Item"]:
        for entry in self.items:
            yield entry[0]
            yield entry[1]


def make_dict(
    args: Union[Tuple[Item, Item, List[Tuple[Item, Item]]], Tuple[Item, Item]]
) -> DictMaker:
    lst = [(args[0], args[1])]
    if len(args) > 2:
        lst += args[2]
    return DictMaker(lst[0][0].start, lst[-1][1].end, lst)


def make_empty_dict(args: Tuple[Token, Token]) -> DictMaker:
    return DictMaker(args[0].start, args[0].end, [])


@dataclasses.dataclass(frozen=True)
class TernaryOp(Item):
    left: Item
    condition: Item
    right: Item

    async def eval(self, root: RootABC) -> TypeT:
        if await self.condition.eval(root):
            return await self.left.eval(root)
        else:
            return await self.right.eval(root)

    def child_items(self) -> Iterable["Item"]:
        return [self.left, self.condition, self.right]


def make_if_else(args: Tuple[Item, Token, Item, Token, Item]) -> TernaryOp:
    return TernaryOp(
        args[0].start,
        args[1].end,
        left=args[0],
        condition=args[2],
        right=args[4],
    )


def a(value: str) -> "Parser[Token, Token]":
    """Eq(a) -> Parser(a, a)

    Returns a parser that parses a token that is equal to the value value.
    """

    def _is_value_eq(token: Token) -> bool:
        return token.value == value

    return some(_is_value_eq).named(f'(a "{value}")')


DOT: Final = skip(a("."))
COMMA: Final = skip(a(","))
COLON: Final = skip(a(":"))

OPEN_TMPL: Final = skip(a("${{"))
CLOSE_TMPL: Final = skip(a("}}"))
OPEN_TMPL2: Final = skip(a("$[["))
CLOSE_TMPL2: Final = skip(a("]]"))

LPAR: Final = skip(a("("))
RPAR = skip(a(")"))

LSQB: Final = skip(a("["))
RSQB = skip(a("]"))

LBRACE: Final = skip(a("{"))
RBRACE = skip(a("}"))

OP_PLUS = a("+")
OP_MINUS = a("-")
OP_MUL = a("*")
OP_DIV = a("/")
OP_MOD = a("%")
OP_BITWISE_OR = a("|")
OP_CMP = a("==") | a("!=") | a("<") | a("<=") | a(">") | a(">=")
OP_NOT = a("not")
OP_OR = a("or")
OP_AND = a("and")

REAL: Final = literal("REAL") | literal("EXP")

INT: Final = literal("INT") | literal("HEX") | literal("OCT") | literal("BIN")

BOOL: Final = literal("BOOL")

STR: Final = literal("STR")

NONE: Final = literal("NONE")

LITERAL: Final = NONE | BOOL | REAL | INT | STR

NAME: Final = some(make_toktype_predicate("NAME"))

LIST_MAKER: Final = forward_decl()

DICT_MAKER: Final = forward_decl()

ATOM: Final = LITERAL | LIST_MAKER | DICT_MAKER

EXPR: Final = forward_decl()

ATOM_EXPR: Final = forward_decl()

LOOKUP_ATTR: Final = DOT + NAME >> lookup_attr

LOOKUP_ITEM: Final = LSQB + EXPR + RSQB >> lookup_item

TRAILER: Final = many(LOOKUP_ATTR | LOOKUP_ITEM)

LOOKUP: Final = NAME + TRAILER >> make_lookup

FUNC_ARGS: Final = maybe(EXPR + many(COMMA + EXPR)) >> make_args

FUNC_CALL: Final = (NAME + LPAR + FUNC_ARGS + RPAR + TRAILER) >> make_call

ATOM_EXPR.define(LPAR + EXPR + RPAR | ATOM | FUNC_CALL | LOOKUP)

# Here we define operator precedence:
# +, - (unary)
# *, /
# +, -
# |
# ==, !=, <, >, <=, =>
# and
# or

# For operators, LL1 grammar rules are used to speed-up parsing

DISJUNCTION: Final = forward_decl()
DISJUNCTION_TRAILER: Final = forward_decl()
CONJUNCTION: Final = forward_decl()
CONJUNCTION_TRAILER: Final = forward_decl()
INVERSION: Final = forward_decl()
COMPARISON: Final = forward_decl()
COMPARISON_TRAILER: Final = forward_decl()
BITWISE_OR: Final = forward_decl()
BITWISE_OR_TRAILER: Final = forward_decl()
SUM: Final = forward_decl()
SUM_TRAILER: Final = forward_decl()
TERM: Final = forward_decl()
TERM_TRAILER: Final = forward_decl()
FACTOR: Final = forward_decl()


DISJUNCTION_TRAILER.define(
    maybe(OP_OR + CONJUNCTION + DISJUNCTION_TRAILER) >> make_op_trailer
)
DISJUNCTION.define(CONJUNCTION + DISJUNCTION_TRAILER >> make_bin_op_expr)

CONJUNCTION_TRAILER.define(
    maybe(OP_AND + INVERSION + CONJUNCTION_TRAILER) >> make_op_trailer
)
CONJUNCTION.define(INVERSION + CONJUNCTION_TRAILER >> make_bin_op_expr)

INVERSION.define(OP_NOT + INVERSION >> make_unary_op_expr | COMPARISON)

COMPARISON_TRAILER.define(
    maybe(OP_CMP + BITWISE_OR + COMPARISON_TRAILER) >> make_op_trailer
)
COMPARISON.define(BITWISE_OR + COMPARISON_TRAILER >> make_bin_op_expr)

BITWISE_OR_TRAILER.define(
    maybe(OP_BITWISE_OR + SUM + BITWISE_OR_TRAILER) >> make_op_trailer
)
BITWISE_OR.define(SUM + BITWISE_OR_TRAILER >> make_bin_op_expr)

SUM_TRAILER.define(maybe((OP_PLUS | OP_MINUS) + TERM + SUM_TRAILER) >> make_op_trailer)
SUM.define(TERM + SUM_TRAILER >> make_bin_op_expr)

TERM_TRAILER.define(
    maybe((OP_MUL | OP_DIV | OP_MOD) + FACTOR + TERM_TRAILER) >> make_op_trailer
)
TERM.define(FACTOR + TERM_TRAILER >> make_bin_op_expr)

FACTOR.define((OP_PLUS | OP_MINUS) + FACTOR >> make_unary_op_expr | ATOM_EXPR)

EXPR.define(
    DISJUNCTION + a("if") + DISJUNCTION + a("else") + EXPR >> make_if_else | DISJUNCTION
)

LIST_MAKER.define(
    (LSQB + EXPR + many(COMMA + EXPR) + maybe(COMMA) + RSQB) >> make_list
    | (LSQB + EXPR + a("for") + NAME + a("in") + EXPR + RSQB) >> make_list_comp
    | (LSQB + EXPR + a("for") + NAME + a("in") + EXPR + a("if") + EXPR + RSQB)
    >> make_list_comp_with_if
    | (a("[") + a("]")) >> make_empty_list
)

DICT_MAKER.define(
    (
        LBRACE
        + EXPR
        + COLON
        + EXPR
        + many(COMMA + EXPR + COLON + EXPR)
        + maybe(COMMA)
        + RBRACE
    )
    >> make_dict  # type: ignore
    | (a("{") + a("}")) >> make_empty_dict
)

TMPL: Final = (OPEN_TMPL + EXPR + CLOSE_TMPL) | (OPEN_TMPL2 + EXPR + CLOSE_TMPL2)

TEXT: Final = some(make_toktype_predicate("TEXT")) >> make_text

PARSER: Final = oneplus(TMPL | TEXT) + skip(finished)


class BaseExpr(Generic[_T], abc.ABC):
    @abc.abstractmethod
    async def eval(self, root: RootABC) -> Optional[_T]:
        pass


IMPLICIT_STR_CONCAT: Final[Tuple[type, ...]] = (str, RemotePath, LocalPath, URL)


class Expr(BaseExpr[_T]):
    allow_none: ClassVar[bool] = True
    allow_expr: ClassVar[bool] = True
    allow_implicit_concat: ClassVar[bool] = True
    type_name: ClassVar[str]
    start: Pos
    end: Pos
    _ret: Union[None, _T]
    _pattern: Union[None, str, _T]
    _parsed: Optional[Sequence[Item]]

    @abc.abstractmethod
    def convert(self, arg: TypeT) -> _T:
        pass

    def _try_convert(self, arg: TypeT, start: Pos, end: Pos) -> None:
        try:
            self._ret = self.convert(arg)
        except (TypeError, ValueError) as exc:
            raise EvalError(str(exc), start, end)

    def __init__(self, start: Pos, end: Pos, pattern: Union[None, str, _T]) -> None:
        self.start = start
        self.end = end
        self._pattern = pattern
        # precalculated value for constant string, allows raising errors earlier
        self._ret = None
        if pattern is not None:
            if isinstance(pattern, str):
                # parse later
                pass
            else:
                # explicit non-string value is passed
                self._try_convert(pattern, start, end)
                return
            tokens = list(tokenize(pattern, start=start))
            if tokens:
                self._parsed = PARSER.parse(tokens)
                if (
                    not self.allow_implicit_concat
                    and self._parsed
                    and len(self._parsed) > 1
                ):
                    raise EvalError(
                        "Implicit concatenation is not allowed for "
                        f"{self.type_name}",
                        start,
                        end,
                    )
            else:
                if not self.allow_implicit_concat:
                    raise EvalError(
                        f"Empty value is not allowed for {self.type_name}",
                        start,
                        end,
                    )
                self._parsed = [Text(start, end, "")]
            assert self._parsed
            if len(self._parsed) == 1 and isinstance(self._parsed[0], Text):
                self._try_convert(self._parsed[0].arg, start, end)
            elif not self.allow_expr:
                raise EvalError(f"Expressions are not allowed in {pattern}", start, end)
        elif self.allow_none:
            self._parsed = None
        else:
            raise EvalError("None is not allowed", start, end)

    @property
    def pattern(self) -> Optional[str]:
        if self._pattern is None:
            return None
        return str(self._pattern)

    @property
    def value(self) -> Optional[_T]:
        if self._ret is None:
            return None
        return self._ret

    async def eval(self, root: RootABC) -> Optional[_T]:
        if self._ret is not None:
            return self._ret
        if self._parsed is not None:
            ret: List[TypeT] = []
            for part in self._parsed:
                try:
                    val = await part.eval(root)
                except asyncio.CancelledError:
                    raise
                except EvalError:
                    raise
                except Exception as exc:
                    raise EvalError(str(exc), part.start, part.end) from exc
                # TODO: add str() function, raise an explicit error if
                # an expresion evaluates non-str type
                # assert isinstance(val, str), repr(val)
                ret.append(val)
            try:
                if len(ret) == 1:
                    return self.convert(ret[0])
                else:
                    assert self.allow_implicit_concat
                    return self.convert("".join(str(item) for item in ret))
            except asyncio.CancelledError:
                raise
            except EvalError:
                raise
            except Exception as exc:
                raise EvalError(str(exc), self._parsed[0].start, self._parsed[-1].end)
        else:
            # __init__() makes sure that the pattern is not None if
            # self.allow_none is False, the check is present here
            # for better readability.
            assert self.allow_none
            return None

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}({self._pattern!r})"

    def __eq__(self, other: Any) -> bool:
        if type(self) is not type(other):
            return False
        assert isinstance(other, self.__class__)
        return self._pattern == other._pattern

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self._pattern))


class StrictExpr(Expr[_T], ABC):
    allow_none = False

    async def eval(self, root: RootABC) -> _T:
        ret = await super().eval(root)
        assert ret is not None
        return ret


class NoConcatMixin:
    allow_implicit_concat = False


class NoExprMixin:
    allow_expr = False


# These comprehensive specializations exist mainly for static type checker


class PrimitiveExprMixin:
    type_name: ClassVar[str] = "primitive"

    def convert(self, arg: TypeT) -> Union[int, bool, float, str]:
        if isinstance(arg, (int, bool, float)):
            return arg
        return str(arg)


class PrimitiveExpr(PrimitiveExprMixin, StrictExpr[Union[int, bool, float, str]]):
    pass


class OptPrimitiveExpr(PrimitiveExprMixin, Expr[Union[int, bool, float, str]]):
    pass


class SimplePrimitiveExpr(
    PrimitiveExprMixin, NoExprMixin, StrictExpr[Union[int, bool, float, str]]
):
    pass


class SimpleOptPrimitiveExpr(
    PrimitiveExprMixin, NoExprMixin, Expr[Union[int, bool, float, str]]
):
    pass


class StrExprMixin:
    type_name: ClassVar[str] = "str"

    def convert(self, arg: TypeT) -> str:
        return str(arg)


class StrExpr(StrExprMixin, StrictExpr[str]):
    pass


class OptStrExpr(StrExprMixin, Expr[str]):
    pass


class SimpleStrExpr(StrExprMixin, NoExprMixin, StrictExpr[str]):
    pass


class SimpleOptStrExpr(StrExprMixin, NoExprMixin, Expr[str]):
    pass


class IdExprMixin:
    type_name: ClassVar[str] = "id"

    def convert(self, arg: TypeT) -> str:
        if not isinstance(arg, str):
            raise TypeError(f"{arg!r} is not a string")
        if not arg.isidentifier():
            raise ValueError(f"{arg!r} is not identifier")
        if arg == arg.upper():
            raise ValueError(
                f"{arg!r} is invalid identifier, "
                "uppercase names are reserved for internal usage"
            )
        return arg


class IdExpr(IdExprMixin, StrictExpr[str]):
    pass


class OptIdExpr(IdExprMixin, Expr[str]):
    pass


class SimpleIdExpr(IdExprMixin, NoExprMixin, StrictExpr[str]):
    pass


class SimpleOptIdExpr(IdExprMixin, NoExprMixin, Expr[str]):
    pass


class URIExprMixin:
    type_name: ClassVar[str] = "URL"

    def convert(self, arg: TypeT) -> URL:
        return URL(arg)  # type: ignore[arg-type]


class URIExpr(URIExprMixin, StrictExpr[URL]):
    pass


class PlatformResourceURIExpr(URIExpr):
    async def eval(self, root: RootABC) -> URL:
        ret = await super().eval(root)
        try:
            # Hack to get flow's project ctx -> project_name
            project_ctx = root.lookup("project")
        except LookupError:
            project_ctx = None

        if project_ctx:
            async with root.client() as cl:
                ret = cl.parse.str_to_uri(
                    str(ret),
                    project_name=project_ctx.project_name,  # type: ignore
                    short=True,
                )
        return ret


async def project_image_ref(img_ref: str, root: RootABC) -> str:
    project_ctx = None
    if img_ref.startswith("image:"):
        try:
            # Hack to get flow's project ctx -> project_name
            project_ctx = root.lookup("project")
        except LookupError:
            pass
    if project_ctx:
        async with root.client() as cl:
            uri = cl.parse.str_to_uri(
                img_ref,
                project_name=project_ctx.project_name,  # type: ignore
                short=True,
            )
            img_ref = str(uri)
    return img_ref


class OptImageRefStrExpr(OptStrExpr):
    async def eval(self, root: RootABC) -> Optional[str]:
        img_ref = await super().eval(root)
        if img_ref is not None:
            return await project_image_ref(img_ref, root)
        return None


class ImageRefStrExpr(StrExpr):
    async def eval(self, root: RootABC) -> str:
        img_ref = await super().eval(root)
        return await project_image_ref(img_ref, root)


class OptURIExpr(URIExprMixin, Expr[URL]):
    pass


class BoolExprMixin(NoConcatMixin):
    type_name: ClassVar[str] = "bool"

    def convert(self, arg: TypeT) -> bool:
        return bool(arg)


class BoolExpr(BoolExprMixin, StrictExpr[bool]):
    pass


class OptBoolExpr(BoolExprMixin, Expr[bool]):
    pass


class SimpleBoolExpr(BoolExprMixin, NoExprMixin, StrictExpr[bool]):
    pass


class SimpleOptBoolExpr(BoolExprMixin, NoExprMixin, Expr[bool]):
    pass


class EnableExprMixin(NoConcatMixin):
    type_name: ClassVar[str] = "enabled_type"

    def convert(self, arg: TypeT) -> Union[bool, AlwaysT]:
        if isinstance(arg, AlwaysT):
            return arg
        if arg == "always()":
            return AlwaysT()
        return bool(arg)


class EnableExpr(EnableExprMixin, StrictExpr[Union[bool, AlwaysT]]):
    pass


class OptEnableExpr(EnableExprMixin, Expr[Union[bool, AlwaysT]]):
    pass


class IntExprMixin(NoConcatMixin):
    type_name: ClassVar[str] = "int"

    def convert(self, arg: TypeT) -> int:
        return int(arg)  # type: ignore[arg-type]


class IntExpr(IntExprMixin, StrictExpr[int]):
    pass


class OptIntExpr(IntExprMixin, Expr[int]):
    pass


class FloatExprMixin(NoConcatMixin):
    type_name: ClassVar[str] = "float"

    def convert(self, arg: TypeT) -> float:
        return float(arg)  # type: ignore[arg-type]


class FloatExpr(FloatExprMixin, StrictExpr[float]):
    pass


class OptFloatExpr(FloatExprMixin, Expr[float]):
    pass


class OptTimeDeltaExpr(OptFloatExpr):
    type_name: ClassVar[str] = "timedelta"
    RE = re.compile(r"^((?P<d>\d+)d)?((?P<h>\d+)h)?((?P<m>\d+)m)?((?P<s>\d+)s)?$")

    def convert(self, arg: TypeT) -> float:
        try:
            return super().convert(arg)
        except (ValueError, SyntaxError):
            assert isinstance(arg, str)
            match = self.RE.match(arg)
            if match is None:
                raise ValueError(f"{arg!r} is not a time delta unit")
            td = datetime.timedelta(
                days=int(match.group("d") or 0),
                hours=int(match.group("h") or 0),
                minutes=int(match.group("m") or 0),
                seconds=int(match.group("s") or 0),
            )
            return td.total_seconds()


class LocalPathMixin:
    type_name: ClassVar[str] = "LocalPath"

    def convert(self, arg: TypeT) -> LocalPath:
        return LocalPath(arg)  # type: ignore[arg-type]


class LocalPathExpr(LocalPathMixin, StrictExpr[LocalPath]):
    pass


class OptLocalPathExpr(LocalPathMixin, Expr[LocalPath]):
    pass


class RemotePathMixin:
    type_name: ClassVar[str] = "RemotePath"

    def convert(self, arg: TypeT) -> RemotePath:
        return RemotePath(arg)  # type: ignore[arg-type]


class RemotePathExpr(RemotePathMixin, StrictExpr[RemotePath]):
    pass


class OptRemotePathExpr(RemotePathMixin, Expr[RemotePath]):
    pass


class OptBashExpr(OptStrExpr):
    def convert(self, arg: TypeT) -> str:
        ret = " ".join(["bash", "-euo", "pipefail", "-c", shlex.quote(str(arg))])
        return ret


class OptPythonExpr(OptStrExpr):
    def convert(self, arg: TypeT) -> str:
        ret = " ".join(["python3", "-uc", shlex.quote(str(arg))])
        return ret


PORT_PAIR_RE = re.compile(r"^\d+:\d+$")


def port_pair_item(arg: TypeT) -> str:
    sarg = str(arg)
    match = PORT_PAIR_RE.match(sarg)
    if match is None:
        raise ValueError(f"{arg!r} is not a LOCAL:REMOTE ports pair")
    return sarg


class PortPairExpr(StrExpr):
    type_name: ClassVar[str] = "PortPair"

    def convert(self, arg: TypeT) -> str:
        return port_pair_item(arg)


class BaseSequenceExpr(BaseExpr[SequenceT], abc.ABC):
    pass


class ConcatSequenceExpr(BaseSequenceExpr):
    def __init__(self, *seqs: BaseSequenceExpr):
        self._seqs = seqs

    async def eval(self, root: RootABC) -> SequenceT:
        ret: List[TypeT] = []
        for seq in self._seqs:
            value = await seq.eval(root)
            if value:
                ret.extend(value)
        return cast(SequenceT, ret)


class SequenceExpr(BaseSequenceExpr, NoConcatMixin, Expr[SequenceT]):
    type_name: ClassVar[str] = "Sequence"

    def __init__(
        self,
        start: Pos,
        end: Pos,
        pattern: Union[None, str, SequenceT],
        item_factory: Callable[[TypeT], TypeT],
    ) -> None:
        super().__init__(start, end, pattern)
        self._item_factory = item_factory

    def convert(self, arg: TypeT) -> SequenceT:
        if not isinstance(arg, SequenceT):
            raise TypeError(f"{arg!r} is not a sequence")
        ret: List[TypeT] = []
        for item in arg:
            ret.append(self._item_factory(item))
        return cast(SequenceT, ret)


class SequenceItemsExpr(BaseSequenceExpr, Generic[_IT]):
    def __init__(self, items: Sequence[Expr[_IT]]) -> None:
        self._items = items

    async def eval(self, root: RootABC) -> SequenceT:
        ret = []
        for item in self._items:
            ret.append(await item.eval(root))
        return cast(SequenceT, ret)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}({self._items!r})"

    def __eq__(self, other: Any) -> bool:
        if type(self) is not type(other):
            return False
        assert isinstance(other, self.__class__)
        return self._items == other._items

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self._items))


class BaseMappingExpr(BaseExpr[MappingT]):
    pass


class MergeMappingsExpr(BaseMappingExpr):
    def __init__(self, *mappings: BaseMappingExpr):
        self._mappings = mappings

    async def eval(self, root: RootABC) -> MappingT:
        ret = {}
        for mapping in reversed(self._mappings):
            value = await mapping.eval(root)
            if value:
                for key in value:
                    ret[key] = value[key]
        return cast(MappingT, ret)


class MappingExpr(BaseMappingExpr, NoConcatMixin, StrictExpr[MappingT]):
    type_name: ClassVar[str] = "Mapping"

    def __init__(
        self,
        start: Pos,
        end: Pos,
        pattern: Union[None, str, MappingT],
        value_factory: Callable[[TypeT], TypeT],
    ) -> None:
        super().__init__(start, end, pattern)
        self._value_factory = value_factory

    def convert(self, arg: TypeT) -> MappingT:
        if not isinstance(arg, MappingT):
            raise TypeError(f"{arg!r} is not a sequence")
        ret = {}
        for key, value in arg.items():  # type: ignore[attr-defined]
            if not isinstance(key, str):
                raise TypeError(f"{key:r} is not a string")
            ret[key] = self._value_factory(value)
        return cast(MappingT, ret)


class MappingItemsExpr(BaseMappingExpr, Generic[_IT]):
    def __init__(self, items: Mapping[str, Expr[_IT]]) -> None:
        self._items = items

    async def eval(self, root: RootABC) -> MappingT:
        ret = {}
        for key, value in self._items.items():
            ret[key] = await value.eval(root)
        return cast(MappingT, ret)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}({self._items!r})"

    def __eq__(self, other: Any) -> bool:
        if type(self) is not type(other):
            return False
        assert isinstance(other, self.__class__)
        return self._items == other._items

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self._items))


def restart_policy_item(arg: TypeT) -> str:
    sarg = str(arg)
    try:
        arg = JobRestartPolicy(sarg)
    except ValueError:
        allowed_values = ", ".join(list(JobRestartPolicy))
        raise ValueError(
            f"{arg!r} is not a valid restart policy, allowed: {allowed_values}"
        )
    return arg


class OptRestartPolicyExpr(OptStrExpr):
    type_name: ClassVar[str] = "RestartPolicy"

    def convert(self, arg: TypeT) -> str:
        return restart_policy_item(arg)



import abc
import apolo_sdk
import click
from apolo_cli.asyncio_utils import Runner
from apolo_sdk import ResourceNotFound
from click.shell_completion import CompletionItem
from contextlib import AsyncExitStack
from typing import Generic, List, Optional, TypeVar, cast

from apolo_flow.batch_runner import BatchRunner
from apolo_flow.cli.root import Root
from apolo_flow.cli.utils import resolve_bake
from apolo_flow.live_runner import LiveRunner
from apolo_flow.storage.api import ApiStorage
from apolo_flow.storage.base import Storage


_T = TypeVar("_T")


class AsyncType(click.ParamType, Generic[_T], abc.ABC):
    def convert(
        self,
        value: str,
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> _T:
        assert ctx is not None
        root = cast(Root, ctx.obj)
        with Runner() as runner:
            return runner.run(self.async_convert(root, value, param, ctx))

    @abc.abstractmethod
    async def async_convert(
        self,
        root: Root,
        value: str,
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> _T:
        pass

    def shell_complete(
        self, ctx: click.Context, param: click.Parameter, incomplete: str
    ) -> List[CompletionItem]:
        root = cast(Root, ctx.obj)
        with Runner() as runner:
            return runner.run(self.async_shell_complete(root, ctx, param, incomplete))

    @abc.abstractmethod
    async def async_shell_complete(
        self, root: Root, ctx: click.Context, param: click.Parameter, incomplete: str
    ) -> List[CompletionItem]:
        pass


class LiveJobType(AsyncType[str]):
    name = "job"

    def __init__(self, allow_all: bool = False):
        self._allow_all = allow_all

    async def async_convert(
        self,
        root: Root,
        value: str,
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> str:
        return value

    async def async_shell_complete(
        self, root: Root, ctx: click.Context, param: click.Parameter, incomplete: str
    ) -> List[CompletionItem]:
        async with AsyncExitStack() as stack:
            client = await stack.enter_async_context(apolo_sdk.get())
            storage: Storage = await stack.enter_async_context(ApiStorage(client))
            runner = await stack.enter_async_context(
                LiveRunner(root.config_dir, root.console, client, storage, root)
            )
            variants = list(runner.flow.job_ids)
            if self._allow_all:
                variants += ["ALL"]
        return [
            CompletionItem(job_id)
            for job_id in variants
            if job_id.startswith(incomplete)
        ]


LIVE_JOB = LiveJobType(allow_all=False)
LIVE_JOB_OR_ALL = LiveJobType(allow_all=True)


class LiveJobSuffixType(AsyncType[str]):
    name = "suffix"

    def __init__(self, *, job_id_param_name: str = "job_id"):
        self._job_id_param_name = job_id_param_name

    async def async_convert(
        self,
        root: Root,
        value: str,
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> str:
        return value

    async def async_shell_complete(
        self, root: Root, ctx: click.Context, param: click.Parameter, incomplete: str
    ) -> List[CompletionItem]:
        job_id = ctx.params[self._job_id_param_name]
        async with AsyncExitStack() as stack:
            client = await stack.enter_async_context(apolo_sdk.get())
            storage: Storage = await stack.enter_async_context(ApiStorage(client))
            runner = await stack.enter_async_context(
                LiveRunner(root.config_dir, root.console, client, storage, root)
            )
            variants = await runner.list_suffixes(job_id)
        return [
            CompletionItem(suffix)
            for suffix in variants
            if suffix.startswith(incomplete)
        ]


class LiveImageType(AsyncType[str]):
    name = "image"

    def __init__(self, allow_all: bool = False):
        self._allow_all = allow_all

    async def async_convert(
        self,
        root: Root,
        value: str,
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> str:
        return value

    async def async_shell_complete(
        self, root: Root, ctx: click.Context, param: click.Parameter, incomplete: str
    ) -> List[CompletionItem]:
        async with AsyncExitStack() as stack:
            client = await stack.enter_async_context(apolo_sdk.get())
            storage: Storage = await stack.enter_async_context(ApiStorage(client))
            runner = await stack.enter_async_context(
                LiveRunner(root.config_dir, root.console, client, storage, root)
            )
            variants = [
                image
                for image, image_ctx in runner.flow.images.items()
                if image_ctx.context is not None
            ]
            if self._allow_all:
                variants += ["ALL"]
        return [
            CompletionItem(image) for image in variants if image.startswith(incomplete)
        ]


LIVE_IMAGE_OR_ALL = LiveImageType(allow_all=True)


class LiveVolumeType(AsyncType[str]):
    name = "volume"

    def __init__(self, allow_all: bool = False):
        self._allow_all = allow_all

    async def async_convert(
        self,
        root: Root,
        value: str,
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> str:
        return value

    async def async_shell_complete(
        self, root: Root, ctx: click.Context, param: click.Parameter, incomplete: str
    ) -> List[CompletionItem]:
        async with AsyncExitStack() as stack:
            client = await stack.enter_async_context(apolo_sdk.get())
            storage: Storage = await stack.enter_async_context(ApiStorage(client))
            runner = await stack.enter_async_context(
                LiveRunner(root.config_dir, root.console, client, storage, root)
            )
            variants = [
                volume.id
                for volume in runner.flow.volumes.values()
                if volume.local is not None
            ]
            if self._allow_all:
                variants += ["ALL"]
        return [
            CompletionItem(image) for image in variants if image.startswith(incomplete)
        ]


LIVE_VOLUME_OR_ALL = LiveVolumeType(allow_all=True)


class BatchType(AsyncType[str]):
    name = "batch"

    def __init__(self, allow_all: bool = False):
        self._allow_all = allow_all

    async def async_convert(
        self,
        root: Root,
        value: str,
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> str:
        return value

    async def async_shell_complete(
        self, root: Root, ctx: click.Context, param: click.Parameter, incomplete: str
    ) -> List[CompletionItem]:
        variants = []
        for file in root.config_dir.config_dir.rglob("*.yml"):
            # We are not trying to parse properly to allow autocompletion of
            # broken yaml files
            if "batch" in file.read_text():
                variants.append(file.stem)
        if self._allow_all:
            variants += ["ALL"]
        return [
            CompletionItem(batch) for batch in variants if batch.startswith(incomplete)
        ]


BATCH = BatchType(allow_all=False)
BATCH_OR_ALL = BatchType(allow_all=True)


class BakeType(AsyncType[str]):
    name = "bake"

    async def async_convert(
        self,
        root: Root,
        value: str,
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> str:
        return value

    async def async_shell_complete(
        self, root: Root, ctx: click.Context, param: click.Parameter, incomplete: str
    ) -> List[CompletionItem]:
        variants = []
        async with AsyncExitStack() as stack:
            client = await stack.enter_async_context(apolo_sdk.get())
            storage: Storage = await stack.enter_async_context(ApiStorage(client))
            runner: BatchRunner = await stack.enter_async_context(
                BatchRunner(root.config_dir, root.console, client, storage, root)
            )
            try:
                async for bake in runner.get_bakes():
                    variants.append(bake.id)
                    if bake.name is not None:
                        variants.append(bake.name)
            except ValueError:
                pass
        return [
            CompletionItem(bake) for bake in variants if bake.startswith(incomplete)
        ]


BAKE = BakeType()


class BakeTaskType(AsyncType[str]):
    name = "task"

    def __init__(
        self,
        *,
        bake_id_param_name: str = "bake_id",
        attempt_no_param_name: str = "bake_id",
        include_started: bool = True,
        include_finished: bool = True,
    ):
        self._bake_id_param_name = bake_id_param_name
        self._attempt_no_param_name = attempt_no_param_name
        self._include_started = include_started
        self._include_finished = include_finished

    async def async_convert(
        self,
        root: Root,
        value: str,
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> str:
        return value

    async def async_shell_complete(
        self, root: Root, ctx: click.Context, param: click.Parameter, incomplete: str
    ) -> List[CompletionItem]:
        variants: List[str] = []
        bake_id = ctx.params[self._bake_id_param_name]
        attempt_no = ctx.params[self._attempt_no_param_name]
        async with AsyncExitStack() as stack:
            client = await stack.enter_async_context(apolo_sdk.get())
            storage: Storage = await stack.enter_async_context(ApiStorage(client))
            runner: BatchRunner = await stack.enter_async_context(
                BatchRunner(root.config_dir, root.console, client, storage, root)
            )
            try:
                bake_id = await resolve_bake(
                    bake_id, project=runner.project_id, storage=storage
                )
                attempt = await runner.get_bake_attempt(bake_id, attempt_no=attempt_no)
            except ResourceNotFound:
                return []
            tasks = [
                task
                async for task in storage.bake(id=bake_id)
                .attempt(id=attempt.id)
                .list_tasks()
            ]
            if self._include_finished:
                variants.extend(
                    ".".join(task.yaml_id) for task in tasks if task.status.is_finished
                )
            if self._include_started:
                variants.extend(
                    ".".join(task.yaml_id) for task in tasks if task.status.is_finished
                )
        return [
            CompletionItem(task) for task in variants if task.startswith(incomplete)
        ]


class ProjectType(AsyncType[str]):
    name = "project"

    async def async_convert(
        self,
        root: Root,
        value: str,
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> str:
        return value

    async def async_shell_complete(
        self, root: Root, ctx: click.Context, param: click.Parameter, incomplete: str
    ) -> List[CompletionItem]:
        variants = []
        async with AsyncExitStack() as stack:
            client = await stack.enter_async_context(apolo_sdk.get())
            storage: Storage = await stack.enter_async_context(ApiStorage(client))
            try:
                async for project in storage.list_projects():
                    variants.append(project.yaml_id)
            except ValueError:
                pass
        return [
            CompletionItem(yaml_id)
            for yaml_id in variants
            if yaml_id.startswith(incomplete)
        ]


PROJECT = ProjectType()



# Dataclasses
from dataclasses import dataclass, field

import enum
from typing import AbstractSet, Mapping, Optional, Sequence, Type, Union

from .expr import (
    BaseExpr,
    EnableExpr,
    IdExpr,
    ImageRefStrExpr,
    MappingT,
    OptBoolExpr,
    OptIdExpr,
    OptIntExpr,
    OptLocalPathExpr,
    OptRemotePathExpr,
    OptRestartPolicyExpr,
    OptStrExpr,
    OptTimeDeltaExpr,
    PlatformResourceURIExpr,
    PrimitiveExpr,
    RemotePathExpr,
    SequenceT,
    SimpleIdExpr,
    SimpleOptBoolExpr,
    SimpleOptIdExpr,
    SimpleOptPrimitiveExpr,
    SimpleOptStrExpr,
    SimpleStrExpr,
    StrExpr,
)
from .tokenizer import Pos


@dataclass(frozen=True)
class Base:
    _start: Pos
    _end: Pos


@dataclass(frozen=True)
class WithSpecifiedFields(Base):
    _specified_fields: AbstractSet[str]


class CacheStrategy(enum.Enum):
    NONE = "none"
    DEFAULT = "default"
    INHERIT = "inherit"


@dataclass(frozen=True)
class Cache(Base):
    # 'default' for root BatchFlowDefaults,
    # 'inherit' for task definitions and actions
    strategy: Optional[CacheStrategy] = field(metadata={"allow_none": True})
    life_span: OptTimeDeltaExpr
    # TODO: maybe add extra key->value mapping for additional cache keys later


@dataclass(frozen=True)
class Project(Base):
    id: SimpleIdExpr
    project_name: SimpleOptStrExpr  # project name can contain "-"
    owner: SimpleOptStrExpr  # user name can contain "-"
    role: SimpleOptStrExpr

    images: Optional[Mapping[str, "Image"]] = field(metadata={"allow_none": True})
    volumes: Optional[Mapping[str, "Volume"]] = field(metadata={"allow_none": True})
    defaults: Optional["BatchFlowDefaults"] = field(metadata={"allow_none": True})
    mixins: Optional[Mapping[str, "ExecUnitMixin"]] = field(
        metadata={"allow_none": True}
    )


# There are 'batch' for pipelined mode and 'live' for interactive one
# (while 'batches' are technically just non-interactive jobs.


class FlowKind(enum.Enum):
    LIVE = "live"  # interactive mode.
    BATCH = "batch"  # pipelined mode


@dataclass(frozen=True)
class Volume(Base):
    remote: PlatformResourceURIExpr  # remote URI, e.g. storage:folder/subfolder
    mount: RemotePathExpr  # mount path inside container
    local: OptLocalPathExpr
    read_only: OptBoolExpr  # True if mounted in read-only mode, False for read-write


@dataclass(frozen=True)
class Image(Base):
    ref: ImageRefStrExpr  # Image reference, e.g. image:my-proj or neuromation/base@v1.6
    context: OptStrExpr
    dockerfile: OptStrExpr
    build_args: Optional[BaseExpr[SequenceT]] = field(metadata={"allow_none": True})
    env: Optional[BaseExpr[MappingT]] = field(metadata={"allow_none": True})
    volumes: Optional[BaseExpr[SequenceT]] = field(metadata={"allow_none": True})
    build_preset: OptStrExpr
    force_rebuild: OptBoolExpr
    extra_kaniko_args: OptStrExpr


@dataclass(frozen=True)
class ExecUnitMixin(WithSpecifiedFields, Base):
    title: OptStrExpr  # Autocalculated if not passed explicitly
    name: OptStrExpr
    image: OptStrExpr
    preset: OptStrExpr
    schedule_timeout: OptTimeDeltaExpr
    entrypoint: OptStrExpr
    cmd: OptStrExpr
    workdir: OptRemotePathExpr
    env: Optional[BaseExpr[MappingT]] = field(metadata={"allow_none": True})
    volumes: Optional[BaseExpr[SequenceT]] = field(metadata={"allow_none": True})
    tags: Optional[BaseExpr[SequenceT]] = field(metadata={"allow_none": True})
    life_span: OptTimeDeltaExpr
    http_port: OptIntExpr
    http_auth: OptBoolExpr
    pass_config: OptBoolExpr
    mixins: Optional[Sequence[StrExpr]] = field(metadata={"allow_none": True})
    restart: OptRestartPolicyExpr


@dataclass(frozen=True)
class ExecUnit(Base):
    title: OptStrExpr  # Autocalculated if not passed explicitly
    name: OptStrExpr
    image: OptStrExpr
    preset: OptStrExpr
    schedule_timeout: OptTimeDeltaExpr
    entrypoint: OptStrExpr
    cmd: OptStrExpr
    workdir: OptRemotePathExpr
    env: Optional[BaseExpr[MappingT]] = field(metadata={"allow_none": True})
    volumes: Optional[BaseExpr[SequenceT]] = field(metadata={"allow_none": True})
    tags: Optional[BaseExpr[SequenceT]] = field(metadata={"allow_none": True})
    life_span: OptTimeDeltaExpr
    http_port: OptIntExpr
    http_auth: OptBoolExpr
    pass_config: OptBoolExpr
    restart: OptRestartPolicyExpr


@dataclass(frozen=True)
class Matrix(Base):
    # AST class is slightly different from YAML representation,
    # in YAML `products` mapping is embedded into the matrix itself.
    products: Mapping[str, BaseExpr[SequenceT]]
    exclude: Sequence[Mapping[str, PrimitiveExpr]]
    include: Sequence[Mapping[str, PrimitiveExpr]]


@dataclass(frozen=True)
class Strategy(Base):
    matrix: Matrix
    fail_fast: OptBoolExpr
    max_parallel: OptIntExpr


@dataclass(frozen=True)
class Param(Base):
    # Possible params in yaml:
    # params:
    #  name: ~
    #  name: value
    #  name:
    #    default: value
    #    descr: description
    default: OptStrExpr
    descr: OptStrExpr


@dataclass(frozen=True)
class JobBase(Base):
    params: Optional[Mapping[str, Param]] = field(metadata={"allow_none": True})


@dataclass(frozen=True)
class JobMixin(WithSpecifiedFields, Base):
    title: OptStrExpr  # Autocalculated if not passed explicitly
    name: OptStrExpr
    image: OptStrExpr
    preset: OptStrExpr
    schedule_timeout: OptTimeDeltaExpr
    entrypoint: OptStrExpr
    cmd: OptStrExpr
    workdir: OptRemotePathExpr
    env: Optional[BaseExpr[MappingT]] = field(metadata={"allow_none": True})
    volumes: Optional[BaseExpr[SequenceT]] = field(metadata={"allow_none": True})
    tags: Optional[BaseExpr[SequenceT]] = field(metadata={"allow_none": True})
    life_span: OptTimeDeltaExpr
    http_port: OptIntExpr
    http_auth: OptBoolExpr
    pass_config: OptBoolExpr
    detach: OptBoolExpr
    browse: OptBoolExpr
    port_forward: Optional[BaseExpr[SequenceT]] = field(metadata={"allow_none": True})
    multi: SimpleOptBoolExpr
    params: Optional[Mapping[str, Param]] = field(metadata={"allow_none": True})
    mixins: Optional[Sequence[StrExpr]] = field(metadata={"allow_none": True})
    restart: OptRestartPolicyExpr


@dataclass(frozen=True)
class Job(ExecUnit, WithSpecifiedFields, JobBase):
    # Interactive job used by Kind.Live flow

    detach: OptBoolExpr
    browse: OptBoolExpr
    port_forward: Optional[BaseExpr[SequenceT]] = field(metadata={"allow_none": True})
    multi: SimpleOptBoolExpr
    mixins: Optional[Sequence[StrExpr]] = field(metadata={"allow_none": True})


class NeedsLevel(enum.Enum):
    RUNNING = "running"
    COMPLETED = "completed"  # pipelined mode


@dataclass(frozen=True)
class TaskBase(Base):
    id: OptIdExpr

    # A set of steps, used in net mode
    # All steps share the same implicit persistent disk volume

    needs: Optional[Mapping[IdExpr, NeedsLevel]] = field(metadata={"allow_none": True})

    # matrix? Do we need a build matrix? Yes probably.

    strategy: Optional[Strategy] = field(metadata={"allow_none": True})

    # continue_on_error: OptBoolExpr
    enable: EnableExpr = field(metadata={"default_expr": "${{ success() }}"})
    cache: Optional[Cache] = field(metadata={"allow_none": True})


@dataclass(frozen=True)
class Task(ExecUnit, WithSpecifiedFields, TaskBase):
    mixins: Optional[Sequence[StrExpr]] = field(metadata={"allow_none": True})


@dataclass(frozen=True)
class TaskMixin(WithSpecifiedFields, Base):
    title: OptStrExpr
    name: OptStrExpr
    image: OptStrExpr
    preset: OptStrExpr
    schedule_timeout: OptTimeDeltaExpr
    entrypoint: OptStrExpr
    cmd: OptStrExpr
    workdir: OptRemotePathExpr
    env: Optional[BaseExpr[MappingT]] = field(metadata={"allow_none": True})
    volumes: Optional[BaseExpr[SequenceT]] = field(metadata={"allow_none": True})
    tags: Optional[BaseExpr[SequenceT]] = field(metadata={"allow_none": True})
    life_span: OptTimeDeltaExpr
    http_port: OptIntExpr
    http_auth: OptBoolExpr
    pass_config: OptBoolExpr
    needs: Optional[Mapping[IdExpr, NeedsLevel]] = field(metadata={"allow_none": True})
    strategy: Optional[Strategy] = field(metadata={"allow_none": True})
    enable: EnableExpr = field(metadata={"default_expr": "${{ success() }}"})
    cache: Optional[Cache] = field(metadata={"allow_none": True})
    mixins: Optional[Sequence[StrExpr]] = field(metadata={"allow_none": True})
    restart: OptRestartPolicyExpr


@dataclass(frozen=True)
class BaseActionCall(Base):
    action: SimpleStrExpr  # action ref
    args: Optional[Mapping[str, PrimitiveExpr]] = field(metadata={"allow_none": True})


@dataclass(frozen=True)
class BaseModuleCall(Base):
    module: SimpleStrExpr  # action ref
    args: Optional[Mapping[str, PrimitiveExpr]] = field(metadata={"allow_none": True})


@dataclass(frozen=True)
class JobActionCall(BaseActionCall, JobBase):
    pass


@dataclass(frozen=True)
class JobModuleCall(BaseModuleCall, JobBase):
    pass


@dataclass(frozen=True)
class TaskActionCall(BaseActionCall, TaskBase):
    pass


@dataclass(frozen=True)
class TaskModuleCall(BaseModuleCall, TaskBase):
    pass


@dataclass(frozen=True)
class FlowDefaults(WithSpecifiedFields, Base):
    tags: Optional[BaseExpr[SequenceT]] = field(metadata={"allow_none": True})

    env: Optional[BaseExpr[MappingT]] = field(metadata={"allow_none": True})
    volumes: Optional[BaseExpr[SequenceT]] = field(metadata={"allow_none": True})
    workdir: OptRemotePathExpr

    life_span: OptTimeDeltaExpr

    preset: OptStrExpr
    schedule_timeout: OptTimeDeltaExpr


@dataclass(frozen=True)
class BatchFlowDefaults(FlowDefaults):
    fail_fast: OptBoolExpr
    max_parallel: OptIntExpr
    cache: Optional[Cache] = field(metadata={"allow_none": True})


@dataclass(frozen=True)
class BaseFlow(Base):
    kind: FlowKind
    id: SimpleOptIdExpr

    title: SimpleOptStrExpr

    images: Optional[Mapping[str, Image]] = field(metadata={"allow_none": True})
    volumes: Optional[Mapping[str, Volume]] = field(metadata={"allow_none": True})
    defaults: Optional[FlowDefaults] = field(metadata={"allow_none": True})


@dataclass(frozen=True)
class LiveFlow(BaseFlow):
    # self.kind == Kind.Job
    mixins: Optional[Mapping[str, JobMixin]] = field(metadata={"allow_none": True})
    jobs: Mapping[str, Union[Job, JobActionCall, JobModuleCall]]


@dataclass(frozen=True)
class BatchFlow(BaseFlow):
    # self.kind == Kind.Batch
    life_span: OptTimeDeltaExpr = field(metadata={"allow_none": True})
    params: Optional[Mapping[str, Param]] = field(metadata={"allow_none": True})
    mixins: Optional[Mapping[str, TaskMixin]] = field(metadata={"allow_none": True})
    tasks: Sequence[Union[Task, TaskActionCall, TaskModuleCall]]

    defaults: Optional[BatchFlowDefaults] = field(metadata={"allow_none": True})


# Action


class ActionKind(enum.Enum):
    LIVE = "live"  # live composite
    BATCH = "batch"  # batch composite
    STATEFUL = "stateful"  # stateful, can be used in batch flow
    LOCAL = "local"  # runs locally, can be used in batch flow


class InputType(enum.Enum):
    INT = "int"
    FLOAT = "float"
    BOOL = "bool"
    STR = "str"

    def to_type(self) -> Union[Type[str], Type[float], Type[int], Type[bool]]:
        if self.value == "int":
            return int
        elif self.value == "float":
            return float
        elif self.value == "bool":
            return bool
        elif self.value == "str":
            return str
        assert False, "Not reachable"


@dataclass(frozen=True)
class Input(Base):
    descr: SimpleOptStrExpr
    default: SimpleOptPrimitiveExpr
    type: InputType = InputType.STR


@dataclass(frozen=True)
class Output(Base):
    descr: SimpleOptStrExpr
    # TODO: split Output class to BatchOutput with value and an Output without it
    value: OptStrExpr  # valid for BatchAction only


@dataclass(frozen=True)
class BaseAction(Base):
    name: SimpleOptStrExpr
    author: SimpleOptStrExpr
    descr: SimpleOptStrExpr
    inputs: Optional[Mapping[str, Input]] = field(metadata={"allow_none": True})

    kind: ActionKind


@dataclass(frozen=True)
class LiveAction(BaseAction):
    job: Job


@dataclass(frozen=True)
class BatchActionOutputs(Base):
    # AST class is slightly different from YAML representation,
    # in YAML `values` mapping is embedded into the outputs itself.
    values: Optional[Mapping[str, Output]] = field(metadata={"allow_none": True})


@dataclass(frozen=True)
class BatchAction(BaseAction):
    outputs: Optional[BatchActionOutputs] = field(metadata={"allow_none": True})
    cache: Optional[Cache] = field(metadata={"allow_none": True})
    images: Optional[Mapping[str, Image]] = field(metadata={"allow_none": True})

    tasks: Sequence[Union[Task, TaskActionCall, TaskModuleCall]]


@dataclass(frozen=True)
class StatefulAction(BaseAction):
    outputs: Optional[Mapping[str, Output]] = field(metadata={"allow_none": True})
    main: ExecUnit
    post: Optional[ExecUnit] = field(metadata={"allow_none": True})
    post_if: EnableExpr = field(metadata={"default_expr": "${{ always() }}"})


@dataclass(frozen=True)
class LocalAction(BaseAction):
    outputs: Optional[Mapping[str, Output]] = field(metadata={"allow_none": True})
    cmd: StrExpr



import dataclasses

import asyncio
import click
import datetime
import secrets
import shlex
import sys
from apolo_sdk import Client, JobDescription, JobStatus, ResourceNotFound
from rich import box
from rich.console import Console
from rich.table import Table
from types import TracebackType
from typing import (
    AbstractSet,
    AsyncContextManager,
    AsyncIterator,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
)

from .config_loader import LiveLocalCL
from .context import ImageCtx, JobMeta, RunningLiveFlow, UnknownJob, VolumeCtx
from .parser import ConfigDir
from .storage.base import ProjectStorage, Storage
from .types import TaskStatus
from .utils import (
    RUNNING_JOB_STATUSES,
    TERMINATED_JOB_STATUSES,
    GlobalOptions,
    encode_global_options,
    fmt_datetime,
    fmt_id,
    make_cmd_exec,
)


@dataclasses.dataclass(frozen=True)
class JobInfo:
    id: str
    status: JobStatus
    raw_id: Optional[str]  # low-level job id, None for never runned jobs
    tags: Iterable[str]
    when: Optional[datetime.datetime]


class LiveRunner(AsyncContextManager["LiveRunner"]):
    def __init__(
        self,
        config_dir: ConfigDir,
        console: Console,
        client: Client,
        storage: Storage,
        global_options: GlobalOptions,
        dry_run: bool = False,
    ) -> None:
        self._config_dir = config_dir
        self._console = console
        self._config_loader: Optional[LiveLocalCL] = None
        self._flow: Optional[RunningLiveFlow] = None
        self._client = client
        self._storage = storage
        self._project_storage: Optional[ProjectStorage] = None
        self._run_apolo_cli = make_cmd_exec(
            "apolo", global_options=encode_global_options(global_options)
        )
        self._run_extras_cli = make_cmd_exec("apolo-extras")
        self._dry_run = dry_run

    async def init_flow(self) -> None:
        if self._flow is not None:
            return
        self._config_loader = LiveLocalCL(self._config_dir, self._client)
        try:
            self._flow = await RunningLiveFlow.create(
                self._config_loader, dry_run=self._dry_run
            )
        except Exception:
            await self._config_loader.close()
            raise

    async def init_project_storage(self) -> None:
        project_name = self.flow.project.project_name
        if not project_name:
            project_name = self._client.config.project_name_or_raise

        project = await self._storage.get_or_create_project(
            self.flow.project.id,
            owner=self.flow.project.owner,
            project_name=project_name,
            cluster=self._client.config.cluster_name,
            org_name=self._client.config.org_name,
        )
        self._project_storage = self._storage.project(id=project.id)

    async def close(self) -> None:
        if self._config_loader is not None:
            await self._config_loader.close()

    async def __aenter__(self) -> "LiveRunner":
        await self.init_flow()
        await self.init_project_storage()
        return self

    async def __aexit__(
        self,
        exc_typ: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        await self.close()

    @property
    def flow(self) -> RunningLiveFlow:
        assert self._flow is not None
        return self._flow

    @property
    def storage(self) -> ProjectStorage:
        assert self._project_storage is not None
        return self._project_storage

    @property
    def client(self) -> Client:
        assert self._client is not None
        return self._client

    async def _ensure_meta(
        self, job_id: str, suffix: Optional[str], *, skip_check: bool = False
    ) -> JobMeta:
        try:
            meta = await self.flow.get_meta(job_id)
            if meta.multi:
                if suffix is None:
                    if not skip_check:
                        raise click.BadArgumentUsage(
                            f"Please provide a suffix for multi-job {fmt_id(job_id)}"
                            " via -s/--suffix option."
                        )
            else:
                if suffix is not None:
                    raise click.BadArgumentUsage(
                        f"Suffix is not allowed for non-multijob {fmt_id(job_id)}"
                    )
            return meta
        except UnknownJob:
            self._console.print(f"[red]Unknown job [b]{job_id}[/b]")
            jobs_str = ", ".join(self.flow.job_ids)
            self._console.print(f"[dim]Existing jobs: {jobs_str}")
            sys.exit(1)

    async def _resolve_jobs(
        self, meta: JobMeta, suffix: Optional[str]
    ) -> AsyncIterator[JobDescription]:
        found = False
        since = datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(
            days=7
        )
        if meta.multi and not suffix:
            async for job in self.client.jobs.list(
                tags=meta.tags,
                reverse=True,
                statuses=(JobStatus.PENDING, JobStatus.RUNNING),
                project_names=[self.flow.project.project_name],
                org_names=[self.client.config.org_name],
            ):
                found = True
                yield job
            async for job in self.client.jobs.list(
                tags=meta.tags,
                reverse=True,
                statuses=(
                    JobStatus.SUSPENDED,
                    JobStatus.SUCCEEDED,
                    JobStatus.FAILED,
                    JobStatus.CANCELLED,
                ),
                since=since,
                project_names=[self.flow.project.project_name],
                org_names=[self.client.config.org_name],
            ):
                found = True
                yield job
        else:
            tags = list(meta.tags)
            if meta.multi and suffix:
                tags.append(f"multi:{suffix}")
            async for job in self.client.jobs.list(
                tags=tags,
                reverse=True,
                limit=1,
                statuses=(JobStatus.PENDING, JobStatus.RUNNING),
                project_names=[self.flow.project.project_name],
                org_names=[self.client.config.org_name],
            ):
                found = True
                yield job
                return
            async for job in self.client.jobs.list(
                tags=tags,
                reverse=True,
                limit=1,
                statuses=(
                    JobStatus.SUSPENDED,
                    JobStatus.SUCCEEDED,
                    JobStatus.FAILED,
                    JobStatus.CANCELLED,
                ),
                since=since,
                project_names=[self.flow.project.project_name],
                org_names=[self.client.config.org_name],
            ):
                found = True
                yield job
                return
        if not found:
            raise ResourceNotFound

    async def _job_status(self, job_id: str) -> List[JobInfo]:
        meta = await self._ensure_meta(job_id, None, skip_check=True)
        ret = []
        found_suffixes = set()
        try:
            async for descr in self._resolve_jobs(meta, None):
                if descr.status == JobStatus.PENDING:
                    when = descr.history.created_at
                elif descr.status == JobStatus.RUNNING:
                    when = descr.history.started_at
                else:
                    when = descr.history.finished_at
                real_id: Optional[str] = None
                for tag in descr.tags:
                    key, sep, val = tag.partition(":")
                    if sep and key == "multi":
                        if val in found_suffixes:
                            break
                        else:
                            found_suffixes.add(val)
                        real_id = f"{job_id} {val}"
                        break
                else:
                    real_id = job_id
                if real_id is not None:
                    ret.append(
                        JobInfo(real_id, descr.status, descr.id, descr.tags, when)
                    )
        except ResourceNotFound:
            ret.append(JobInfo(job_id, JobStatus.UNKNOWN, None, meta.tags, None))
        return ret

    async def list_suffixes(self, job_id: str) -> AbstractSet[str]:
        meta = await self._ensure_meta(job_id, None, skip_check=True)
        result = set()
        try:
            async for job in self._resolve_jobs(meta, None):
                for tag in job.tags:
                    key, sep, val = tag.partition(":")
                    if sep and key == "multi":
                        result.add(val)
        except ResourceNotFound:
            pass
        return result

    async def ps(self) -> None:
        """Return statuses for all jobs from the flow"""

        # TODO: make concurent queries for job statuses
        loop = asyncio.get_event_loop()
        table = Table(box=box.MINIMAL_HEAVY_HEAD)
        table.add_column("JOB", style="bold")
        table.add_column("STATUS")
        table.add_column("RAW ID", style="bright_black")
        table.add_column("WHEN")

        tasks = []
        for job_id in self.flow.job_ids:
            tasks.append(loop.create_task(self._job_status(job_id)))

        for bulk in await asyncio.gather(*tasks):
            for info in bulk:
                table.add_row(
                    info.id,
                    TaskStatus(info.status),
                    info.raw_id or "N/A",
                    fmt_datetime(info.when),
                )

        self._console.print(table)

    async def status(self, job_id: str, suffix: Optional[str]) -> None:
        meta_ctx = await self._ensure_meta(job_id, suffix)
        try:
            async for descr in self._resolve_jobs(meta_ctx, suffix):
                await self._run_apolo_cli("status", descr.id)
        except ResourceNotFound:
            self._console.print(f"Job [b]{job_id}[/b] is not running")
            sys.exit(1)

    async def _try_attach_to_running(
        self,
        job_id: str,
        suffix: Optional[str],
        args: Optional[Tuple[str]],
        params: Mapping[str, str],
    ) -> bool:
        is_multi = await self.flow.is_multi(job_id)

        if not is_multi or suffix:
            meta = await self._ensure_meta(job_id, suffix)
            try:
                jobs = []
                async for descr in self._resolve_jobs(meta, suffix):
                    jobs.append(descr)
                if len(jobs) > 1:
                    # Should never happen, but just in case
                    raise click.ClickException(
                        f"Found multiple running jobs for id {job_id} and"
                        f" suffix {suffix}:\n"
                        "\n".join(job.id for job in jobs)
                    )
                assert len(jobs) == 1
                descr = jobs[0]
                if is_multi and args:
                    raise click.ClickException(
                        "Multi job with such suffix is already running."
                    )
                if descr.status == JobStatus.PENDING:
                    self._console.print(
                        f"Job [b]{job_id}[/b] is pending, " "try again later"
                    )
                    sys.exit(2)
                if descr.status == JobStatus.RUNNING:
                    self._console.print(
                        f"Job [b]{job_id}[/b] is running, " "connecting..."
                    )
                    if not is_multi:
                        job = await self.flow.get_job(job_id, params)
                    else:
                        assert suffix is not None
                        job = await self.flow.get_multi_job(job_id, suffix, (), params)
                    # attach to job if needed, browse first
                    if job.browse:
                        await self._run_apolo_cli("job", "browse", descr.id)
                    if not job.detach:
                        await self._run_apolo_cli("attach", descr.id)
                    return True
                # Here the status is SUCCEED, CANCELLED or FAILED, restart
            except ResourceNotFound:
                # Job does not exist, run it
                pass
        return False

    async def run(
        self,
        job_id: str,
        suffix: Optional[str],
        args: Optional[Tuple[str]],
        params: Mapping[str, str],
    ) -> None:
        """Run a named job"""
        assert self._flow is not None

        meta_ctx = await self._ensure_meta(job_id, suffix, skip_check=True)
        is_multi = meta_ctx.multi

        if not is_multi and args:
            raise click.BadArgumentUsage(
                "Additional job arguments are supported by multi-jobs only"
            )

        if not self._dry_run and await self._try_attach_to_running(
            job_id, suffix, args, params
        ):
            return  # Attached to running job

        if not is_multi:
            job = await self.flow.get_job(job_id, params)
        else:
            if suffix is None:
                suffix = secrets.token_hex(5)
            job = await self.flow.get_multi_job(job_id, suffix, args, params)

        run_args = ["run"]
        if job.title:
            run_args.append(f"--description={job.title}")
        if job.name:
            name = job.name
        else:
            first_part = self._flow.project.id.replace("_", "-").strip("-")
            while "--" in first_part:
                first_part = first_part.replace("--", "-")
            second_part = job_id
            if is_multi:
                second_part += f"-{suffix}"
            second_part = second_part.replace("_", "-").strip("-")
            while "--" in second_part:
                second_part = second_part.replace("--", "-")

            first_part = first_part[: 40 - len(second_part) - 1]
            name = first_part + "-" + second_part
            while "--" in name:
                name = name.replace("--", "-")
        run_args.append(f"--name={name}")
        run_args.append(f"--project={self.flow.project.project_name}")
        if job.preset is not None:
            run_args.append(f"--preset={job.preset}")
        if job.schedule_timeout is not None:
            run_args.append(f"--schedule-timeout={int(job.schedule_timeout)}s")
        if job.http_port is not None:
            run_args.append(f"--http-port={job.http_port}")
        if job.http_auth is not None:
            if job.http_auth:
                run_args.append(f"--http-auth")
            else:
                run_args.append(f"--no-http-auth")
        if job.entrypoint:
            run_args.append(f"--entrypoint={job.entrypoint}")
        if job.workdir is not None:
            run_args.append(f"--workdir={job.workdir}")
        for k, v in job.env.items():
            run_args.append(f"--env={k}={v}")
        for v in job.volumes:
            run_args.append(f"--volume={v}")
        for t in job.tags:
            run_args.append(f"--tag={t}")
        if job.life_span is not None:
            run_args.append(f"--life-span={int(job.life_span)}s")
        if job.browse:
            run_args.append(f"--browse")
        if job.detach:
            run_args.append(f"--detach")
        for pf in job.port_forward:
            run_args.append(f"--port-forward={pf}")
        if job.pass_config:
            run_args.append(f"--pass-config")
        if job.restart:
            run_args.append(f"--restart={job.restart}")

        run_args.append(job.image)

        run_args.extend(["--"])
        if job.cmd:
            run_args.extend(shlex.split(job.cmd))

        if job.multi and args:
            run_args.extend(args)

        if self._dry_run:
            run_args = ["apolo", *run_args]
            self._console.print(
                " ".join(shlex.quote(arg) for arg in run_args), soft_wrap=True
            )
            return

        live_job_metas = []
        for job_id in self.flow.job_ids:
            job_meta = await self.flow.get_meta(job_id)
            live_job_metas.append(job_meta)
        for job_meta in live_job_metas:
            await self.storage.replace_live_job(
                multi=job_meta.multi,
                yaml_id=job_meta.id,
                tags=job_meta.tags,
            )
        await self._run_apolo_cli(*run_args)

    async def logs(self, job_id: str, suffix: Optional[str]) -> None:
        """Return job logs"""
        meta_ctx = await self._ensure_meta(job_id, suffix)
        try:
            async for descr in self._resolve_jobs(meta_ctx, suffix):
                await self._run_apolo_cli("logs", descr.id)
        except ResourceNotFound:
            self._console.print(f"Job {fmt_id(job_id)} is not running")
            sys.exit(1)

    async def kill_job(self, job_id: str, suffix: Optional[str]) -> bool:
        """Kill named job"""
        meta_ctx = await self._ensure_meta(job_id, suffix)

        try:
            async for descr in self._resolve_jobs(meta_ctx, suffix):
                if descr.status in RUNNING_JOB_STATUSES:
                    await self.client.jobs.kill(descr.id)
                    descr = await self.client.jobs.status(descr.id)
                    while descr.status not in TERMINATED_JOB_STATUSES:
                        await asyncio.sleep(0.2)
                        descr = await self.client.jobs.status(descr.id)
                    return True
        except ResourceNotFound:
            pass
        return False

    async def kill(self, job_id: str, suffix: Optional[str]) -> None:
        """Kill named job"""
        if await self.kill_job(job_id, suffix):
            self._console.print(f"Killed job [b]{job_id}[/b]")
        else:
            self._console.print(f"Job [b]{job_id}[/b] is not running")

    async def kill_all(self) -> None:
        """Kill all jobs"""
        tasks = []
        loop = asyncio.get_event_loop()

        async def kill(descr: JobDescription) -> str:
            tag_dct = {}
            for tag in descr.tags:
                key, sep, val = tag.partition(":")
                if sep:
                    tag_dct[key] = val
            job_id = tag_dct["job"]
            suffix = tag_dct.get("multi")
            await self.client.jobs.kill(descr.id)
            try:
                descr = await self.client.jobs.status(descr.id)
                while descr.status not in TERMINATED_JOB_STATUSES:
                    await asyncio.sleep(0.2)
                    descr = await self.client.jobs.status(descr.id)
            except ResourceNotFound:
                pass
            if suffix:
                return f"{job_id} {suffix}"
            else:
                return job_id

        async for descr in self.client.jobs.list(
            tags=self.flow.tags, statuses=RUNNING_JOB_STATUSES
        ):
            tasks.append(loop.create_task(kill(descr)))

        for job_info in await asyncio.gather(*tasks):
            self._console.print(f"Killed job [b]{job_info}[/b]")

    # volumes subsystem

    async def find_volume(self, volume: str) -> VolumeCtx:
        volume_ctx = self.flow.volumes.get(volume)
        if volume_ctx is None:
            self._console.print(f"[red]Unknown volume [b]{volume}[/b]")
            volumes = sorted(volume for volume in self.flow.volumes.keys())
            volumes_str = ",".join(volumes)
            self._console.print(f"[dim]Existing volumes: {volumes_str}")
            sys.exit(1)
        if volume_ctx.local is None:
            self._console.print(f"Volume's [b]local[/b] part is not set")
            sys.exit(2)
        return volume_ctx

    async def upload(self, volume: str) -> None:
        volume_ctx = await self.find_volume(volume)
        await self._run_apolo_cli(
            "mkdir",
            "--parents",
            str(volume_ctx.remote.parent),
        )
        await self._run_apolo_cli(
            "cp",
            "--recursive",
            "--update",
            "--no-target-directory",
            str(volume_ctx.full_local_path),
            str(volume_ctx.remote),
        )

    async def download(self, volume: str) -> None:
        volume_ctx = await self.find_volume(volume)
        await self._run_apolo_cli(
            "cp",
            "--recursive",
            "--update",
            "--no-target-directory",
            str(volume_ctx.remote),
            str(volume_ctx.full_local_path),
        )

    async def clean(self, volume: str) -> None:
        volume_ctx = await self.find_volume(volume)
        await self._run_apolo_cli("rm", "--recursive", str(volume_ctx.remote))

    async def upload_all(self) -> None:
        for volume in self.flow.volumes.values():
            if volume.local is not None:
                await self.upload(volume.id)

    async def download_all(self) -> None:
        for volume in self.flow.volumes.values():
            if volume.local is not None:
                await self.download(volume.id)

    async def clean_all(self) -> None:
        for volume in self.flow.volumes.values():
            if volume.local is not None:
                await self.clean(volume.id)

    async def mkvolumes(self) -> None:
        for volume in self.flow.volumes.values():
            if volume.local is not None:
                volume_ctx = await self.find_volume(volume.id)
                self._console.print(f"Create volume [b]{volume.id}[/b]")
                await self._run_apolo_cli(
                    "mkdir",
                    "--parents",
                    str(volume_ctx.remote),
                )

    # images subsystem

    async def find_image(self, image: str) -> ImageCtx:
        image_ctx = self.flow.images.get(image)
        if image_ctx is None:
            self._console.print(f"[red]Unknown image [b]{image}[/b]")
            images = sorted(image for image in self.flow.images.keys())
            images_str = ",".join(images)
            self._console.print(f"[dim]Existing images: {images_str}")
            sys.exit(1)
        if image_ctx.context is None:
            self._console.print(f"Image's [b]context[/b] part is not set")
            sys.exit(2)
        if image_ctx.dockerfile is None:
            self._console.print(f"Image's [b]dockerfile[/b] part is not set")
            sys.exit(3)
        return image_ctx

    async def build(self, image: str, force_overwrite: bool) -> None:
        image_ctx = await self.find_image(image)
        cmd = []
        assert image_ctx.dockerfile_rel is not None
        assert image_ctx.context is not None
        cmd.append(f"--project={self.flow.project.project_name}")
        cmd.append(f"--file={image_ctx.dockerfile_rel.as_posix()}")
        for arg in image_ctx.build_args:
            cmd.append(f"--build-arg={arg}")
        for vol in image_ctx.volumes:
            cmd.append(f"--volume={vol}")
        for k, v in image_ctx.env.items():
            cmd.append(f"--env={k}={v}")
        if force_overwrite or image_ctx.force_rebuild:
            cmd.append("--force-overwrite")
        if image_ctx.build_preset is not None:
            cmd.append(f"--preset={image_ctx.build_preset}")
        if image_ctx.extra_kaniko_args:
            cmd.append(
                f"--extra-kaniko-args={shlex.quote(image_ctx.extra_kaniko_args)}"
            )
        cmd.append(str(image_ctx.context))
        cmd.append(str(image_ctx.ref))
        await self._run_extras_cli("image", "build", *cmd)

    async def build_all(self, force_overwrite: bool) -> None:
        for image, image_ctx in self.flow.images.items():
            if image_ctx.context is not None:
                await self.build(image, force_overwrite=force_overwrite)



import dataclasses

import abc
import collections
from abc import abstractmethod
from typing import AbstractSet, Any, Callable, Iterable, List, Optional, Tuple, Type
from typing_extensions import get_type_hints as _get_hints

from apolo_flow.context import Context, ModuleContext, TagsCtx
from apolo_flow.expr import (
    AttrGetter,
    EvalError,
    Expr,
    Item,
    ItemGetter,
    ListCompMaker,
    Lookup,
)


def get_hints(obj: Any) -> Any:
    return _get_hints(obj, include_extras=True)


def validate_expr(
    expr: Expr[Any],
    context: Type[Context],
    known_needs: AbstractSet[str] = frozenset(),
    known_inputs: AbstractSet[str] = frozenset(),
) -> List[EvalError]:
    errors: List[EvalError] = []
    for lookup, local_vars in iter_lookups(expr):
        validate_lookup(
            lookup, context, errors.append, known_needs, known_inputs, local_vars
        )
    return errors


def iter_lookups(expr: Expr[Any]) -> Iterable[Tuple[Lookup, AbstractSet[str]]]:
    def _iter_lookups(
        top_level_items: Iterable[Item], local_vars: AbstractSet[str] = frozenset()
    ) -> Iterable[Tuple[Lookup, AbstractSet[str]]]:
        for item in top_level_items:
            if isinstance(item, Lookup):
                yield item, local_vars
            item_locals = local_vars
            if isinstance(item, ListCompMaker):
                item_locals = local_vars | {item.var_name}
            yield from _iter_lookups(item.child_items(), item_locals)

    return _iter_lookups(getattr(expr, "_parsed", None) or [])


def _get_dataclass_field_type(dataclass: Any, attr: str) -> Optional[Any]:
    if getattr(dataclass, "__origin__", None) is not None:
        # This is generic, probably ModuleContext
        real_class = dataclass.__origin__
        res = _get_dataclass_field_type(real_class, attr)
        if res is None and issubclass(real_class, ModuleContext):
            res = _get_dataclass_field_type(dataclass.__args__[0], attr)
        return res
    res = get_hints(dataclass).get(attr)
    if res is None:
        # Maybe this is a property?
        class_attr = getattr(dataclass, attr, None)
        fget = getattr(class_attr, "fget", None)
        if fget:
            res = get_hints(fget).get("return")
    return res


class GetterVisitor(abc.ABC):
    @abstractmethod
    def dataclass(self, obj: Any) -> Optional[Any]:
        pass

    @abstractmethod
    def mapping(self, type_args: Tuple[Any, ...], metadata: str) -> Optional[Any]:
        pass

    @abstractmethod
    def set(self, type_args: Tuple[Any, ...], metadata: str) -> Optional[Any]:
        pass


def _format_obj_name(obj: Any) -> str:
    if hasattr(obj, "__name__"):
        return str(obj.__name__)
    if obj == TagsCtx:
        return "TagsCtx"
    return str(obj)


class AttrGetterVisitor(GetterVisitor):
    def __init__(
        self,
        record_error: Callable[[EvalError], None],
        known_needs: AbstractSet[str],
        known_inputs: AbstractSet[str],
        getter: AttrGetter,
    ):
        self._record_error = record_error
        self._known_needs = known_needs
        self._known_inputs = known_inputs
        self._getter = getter

    def dataclass(self, obj: Any) -> Optional[Any]:
        new_ctx = _get_dataclass_field_type(obj, self._getter.name)
        if new_ctx is None:
            self._record_error(
                EvalError(
                    f"'{_format_obj_name(obj)}' has no attribute "
                    f"'{self._getter.name}'",
                    self._getter.start,
                    self._getter.end,
                )
            )
        return new_ctx

    def mapping(self, type_args: Tuple[Any, ...], metadata: str) -> Optional[Any]:
        if metadata == "NeedsCtx" and self._getter.name not in self._known_needs:
            self._record_error(
                EvalError(
                    f"Task '{self._getter.name}' is not available under "
                    f"needs context",
                    self._getter.start,
                    self._getter.end,
                )
            )
        if metadata == "InputsCtx" and self._getter.name not in self._known_inputs:
            self._record_error(
                EvalError(
                    f"Input '{self._getter.name}' is undefined",
                    self._getter.start,
                    self._getter.end,
                )
            )
        return type_args[1]

    def set(self, type_args: Tuple[Any, ...], metadata: str) -> Optional[Any]:
        self._record_error(
            EvalError(
                f"'{metadata}' has no attribute " f"'{self._getter.name}'",
                self._getter.start,
                self._getter.end,
            )
        )
        return None


class ItemGetterVisitor(GetterVisitor):
    def __init__(self, record_error: Callable[[EvalError], None], getter: ItemGetter):
        self._record_error = record_error
        self._getter = getter

    def dataclass(self, obj: Any) -> Optional[Any]:
        self._record_error(
            EvalError(
                f"'{_format_obj_name(obj)}' is not subscriptable",
                self._getter.start,
                self._getter.end,
            )
        )
        return None

    def mapping(self, type_args: Tuple[Any, ...], metadata: str) -> Optional[Any]:
        return type_args[1]

    def set(self, type_args: Tuple[Any, ...], metadata: str) -> Optional[Any]:
        self._record_error(
            EvalError(
                f"'{metadata}' is not subscriptable",
                self._getter.start,
                self._getter.end,
            )
        )
        return None


def validate_lookup(
    lookup: Lookup,
    context: Type[Context],
    record_error: Callable[[EvalError], None],
    known_needs: AbstractSet[str],
    known_inputs: AbstractSet[str],
    local_vars: AbstractSet[str],
) -> None:
    if lookup.root in local_vars:
        return
    ctx = _get_dataclass_field_type(context, lookup.root)
    if ctx is None:
        record_error(
            EvalError(f"Unknown context '{lookup.root}'", lookup.start, lookup.end)
        )
        return
    for getter in lookup.trailer:
        if isinstance(getter, AttrGetter):
            visitor: GetterVisitor = AttrGetterVisitor(
                record_error,
                known_needs,
                known_inputs,
                getter,
            )
        else:
            assert isinstance(getter, ItemGetter)
            visitor = ItemGetterVisitor(
                record_error,
                getter,
            )

        if dataclasses.is_dataclass(ctx):
            ctx = visitor.dataclass(ctx)
        elif hasattr(ctx, "__metadata__"):  # This is typing.Annotated
            # The following code is ugly, unfortunately typing got
            # clear retrospection api only recently,
            # and it is not backported to python3.6
            annotation = ctx.__metadata__[0]
            origin_type = ctx.__args__[0]
            type_args = origin_type.__args__
            if hasattr(origin_type, "__extra__"):
                # python 3.6
                origin_abc = origin_type.__extra__
            else:
                origin_abc = origin_type.__origin__

            if origin_abc == collections.abc.Mapping:
                ctx = visitor.mapping(type_args, annotation)
            elif origin_abc == collections.abc.Set:
                ctx = visitor.set(type_args, annotation)
        else:
            ctx = None  # Unknown ...Ctx type, skip check

        if ctx is None:
            return



import dataclasses

import re
from typing import Iterator, Optional

from .types import LocalPath


@dataclasses.dataclass(frozen=True)
class Pos:
    line: int
    col: int
    filename: LocalPath


@dataclasses.dataclass(frozen=True)
class Token:
    type: str
    value: str
    start: Pos
    end: Pos

    def _pos_str(self) -> str:
        sl, sp = self.start.line, self.start.col
        el, ep = self.end.line, self.end.col
        return "%d,%d-%d,%d:" % (sl, sp, el, ep)

    def __str__(self) -> str:
        s = f"{self._pos_str()} {self.type} '{self.value}'"
        return s.strip()

    @property
    def name(self) -> str:
        return self.value

    def pformat(self) -> str:
        return "{} {} '{}'".format(
            self._pos_str().ljust(20),
            self.type.ljust(14),
            self.value,
        )


class LexerError(Exception):
    def __init__(self, place: Pos, msg: str) -> None:
        self.place = place
        self.msg = msg

    def __str__(self) -> str:
        s = "cannot tokenize data"
        return f'{s}: {self.place.line},{self.place.col}: "{self.msg}"'


class Tokenizer:
    TOKENS = [
        ("LTMPL", r"\$\{\{"),
        ("RTMPL", r"\}\}"),
        ("LTMPL2", r"\$\[\["),
        ("RTMPL2", r"\]\]"),
        ("SPACE", r"[ \t]+"),
        ("NONE", r"None"),
        ("BOOL", r"True|False"),
        ("REAL", r"-?[0-9]+\.[0-9]*([Ee][+\-]?[0-9]+)*"),
        ("EXP", r"-?[0-9]+\.[0-9]*([Ee][+\-]?[0-9]+)*[+\-]?e[0-9]+"),
        ("HEX", r"0[xX][0-9a-fA-F_]+"),
        ("OCT", r"0[oO][0-7_]+"),
        ("BIN", r"0[bB][0-1_]+"),
        ("INT", r"-?[0-9][0-9_]*"),
        ("STR", r"'[^']*'|" r'"[^"]*"'),
        ("OP", r"(==|!=|or|and|<=|<|>=|>|\||\+|-|\*|/|%)(?=[ \t])"),
        ("UNARY_OP", r"(not)(?=[ \t])"),
        ("NAME", r"[A-Za-z][A-Za-z_0-9]*"),
        ("DOT", r"\."),
        ("COMMA", r","),
        ("COLON", r":"),
        ("LPAR", r"\("),
        ("RPAR", r"\)"),
        ("LSQB", r"\["),
        ("RSQB", r"\]"),
        ("LBRACE", r"\{"),
        ("RBRACE", r"\}"),
    ]
    TOKENS_RE = re.compile("|".join(f"(?P<{typ}>{regexp})" for typ, regexp in TOKENS))

    LTMPL_RE = re.compile(r"\$\{\{|\$\[\[")
    RTMPL_RE = re.compile(r"\}\}|\]\]")

    def make_token(self, typ: str, value: str, pos: Pos) -> Token:
        nls = value.count("\n")
        n_line = pos.line + nls
        if nls == 0:
            n_col = pos.col + len(value)
        else:
            n_col = len(value) - value.rfind("\n") - 1
        return Token(typ, value, pos, Pos(n_line, n_col, pos.filename))

    def match_specs(self, s: str, i: int, start: Pos, pos: Pos) -> Token:
        m = self.TOKENS_RE.match(s, i)
        if m is not None:
            assert m.lastgroup
            return self.make_token(m.lastgroup, m.group(), pos)
        else:
            errline = s.splitlines()[pos.line - start.line]
            raise LexerError(pos, " " * start.col + errline)

    def match_text(self, s: str, i: int, start: Pos, pos: Pos) -> Optional[Token]:
        ltmpl = self.LTMPL_RE.search(s, i)
        if ltmpl is None:
            idx = len(s)
        else:
            if ltmpl.start() == i:
                return None
            else:
                idx = ltmpl.start()
        return self.make_token("TEXT", s[i:idx], pos)

    def __call__(self, s: str, start: Pos) -> Iterator[Token]:
        in_expr = False
        length = len(s)
        pos = start
        i = 0
        while i < length:
            if in_expr:
                t = self.match_specs(s, i, start, pos)
                if t.type in ("RTMPL", "RTMPL2"):
                    in_expr = False
            else:
                txt = self.match_text(s, i, start, pos)
                if txt is None:
                    in_expr = True
                    t = self.match_specs(s, i, start, pos)
                else:
                    t = txt
            if t.type != "SPACE":
                yield t
            pos = t.end
            i += len(t.value)


tokenize = Tokenizer()



from dataclasses import dataclass

import enum
from pathlib import Path, PurePosixPath
from typing import ClassVar, List, Sequence, Tuple


LocalPath = Path
RemotePath = PurePosixPath
FullID = Tuple[str, ...]


class AlwaysT:
    instance: ClassVar["AlwaysT"]

    def __str__(self) -> str:
        return "always()"

    def __new__(cls) -> "AlwaysT":
        if not hasattr(cls, "instance"):
            cls.instance = super().__new__(cls)
        return cls.instance


class TaskStatus(str, enum.Enum):
    # Almost copy of apolo_sdk.JobStatus, but adds new SKIPPED state

    def __rich__(self) -> str:
        return f"[{COLORS[self]}]{self.value}"

    UNKNOWN = "unknown"
    PENDING = "pending"
    RUNNING = "running"
    SUCCEEDED = "succeeded"
    FAILED = "failed"
    CANCELLED = "cancelled"
    SKIPPED = "skipped"
    CACHED = "cached"

    @property
    def is_pending(self) -> bool:
        return self == self.PENDING

    @property
    def is_running(self) -> bool:
        return self == self.RUNNING

    @property
    def is_finished(self) -> bool:
        return self in (
            self.SUCCEEDED,
            self.FAILED,
            self.CANCELLED,
            self.SKIPPED,
            self.CACHED,
        )

    @classmethod
    def values(cls) -> List[str]:
        return [item.value for item in cls]

    @classmethod
    def active_values(cls) -> List[str]:
        return [item.value for item in cls if not item.is_finished]

    @classmethod
    def finished_values(cls) -> List[str]:
        return [item.value for item in cls if item.is_finished]


COLORS = {
    TaskStatus.UNKNOWN: "reverse bright_black",
    TaskStatus.PENDING: "cyan",
    TaskStatus.RUNNING: "blue",
    TaskStatus.SUCCEEDED: "green",
    TaskStatus.FAILED: "red",
    TaskStatus.CANCELLED: "yellow",
    TaskStatus.SKIPPED: "bright_black",
    TaskStatus.CACHED: "magenta",
}


class ImageStatus(str, enum.Enum):
    PENDING = "pending"
    BUILDING = "building"
    BUILT = "built"
    BUILD_FAILED = "build_failed"
    CACHED = "cached"

    def __rich__(self) -> str:
        return f"[{IMAGE_STATUS_COLORS[self]}]{self.value}"


IMAGE_STATUS_COLORS = {
    ImageStatus.PENDING: "cyan",
    ImageStatus.BUILDING: "blue",
    ImageStatus.BUILT: "green",
    ImageStatus.BUILD_FAILED: "red",
    ImageStatus.CACHED: "magenta",
}


@dataclass(frozen=True)
class GitInfo:
    sha: str
    branch: str
    tags: Sequence[str]



import click
import os
import sys

from apolo_flow.cli.utils import wrap_async
from apolo_flow.types import LocalPath

from .root import Root


CFG_FILE = {"bash": LocalPath("~/.bashrc"), "zsh": LocalPath("~/.zshrc")}
SOURCE_CMD = {"bash": "bash_source", "zsh": "zsh_source"}

ACTIVATION_TEMPLATE = 'eval "$(_APOLO_FLOW_COMPLETE={cmd} {exe})"'


@click.group()
def completion() -> None:
    """
    Output shell completion code.
    """


@click.command()
@click.argument("shell", type=click.Choice(["bash", "zsh"]))
@wrap_async()
async def generate(root: Root, shell: str) -> None:
    """
    Provide instruction for shell completion generation.
    """
    root.console.print(f"Push the following line into your {CFG_FILE[shell]}")
    root.console.print(
        ACTIVATION_TEMPLATE.format(cmd=SOURCE_CMD[shell], exe=sys.argv[0])
    )


@click.command()
@click.argument("shell", type=click.Choice(["bash", "zsh"]))
@wrap_async()
async def patch(root: Root, shell: str) -> None:
    """
    Automatically patch shell configuration profile to enable completion
    """
    GENERATED_START = (
        b"# Start of generated by apolo-flow. Please do not edit this comment.\n"
    )
    GENERATED_END = (
        b"\n# End of generated by apolo-flow. Please do not edit this comment.\n"
    )

    profile_file = CFG_FILE[shell].expanduser()

    code = (
        GENERATED_START
        + os.fsencode(
            ACTIVATION_TEMPLATE.format(cmd=SOURCE_CMD[shell], exe=sys.argv[0])
        )
        + GENERATED_END
    )

    try:
        with profile_file.open("rb+") as profile:
            content = profile.read()
    except FileNotFoundError:
        content = b""

    start = content.find(GENERATED_START)
    if start != -1:
        end = content.find(GENERATED_END)
        if end == -1:
            raise click.ClickException(
                f"Malformed guarding comments. Please edit {profile_file} manually"
            )
        content = content[:start] + code + content[end + len(GENERATED_END) :]
    else:
        if content != b"":
            content += b"\n" + code
        else:
            content = code

    with profile_file.open("wb+") as profile:
        profile.write(content)
    root.console.print(f"Added completion configuration into '{profile_file}'")


completion.add_command(generate)
completion.add_command(patch)



import click
import functools
import re
import sys
from apolo_cli.asyncio_utils import Runner
from apolo_sdk import ResourceNotFound
from asyncio import iscoroutinefunction
from click.types import convert_type
from contextlib import contextmanager
from typing import Any, Awaitable, Callable, Iterator, TypeVar
from typing_extensions import ParamSpec

from apolo_flow.storage.base import Storage


@contextmanager
def _runner() -> Iterator[Runner]:
    # Suppress prints unhandled exceptions
    # on event loop closing to overcome aiohttp bug
    # Check https://github.com/aio-libs/aiohttp/issues/4324
    try:
        with Runner() as runner:
            yield runner
            sys.stderr = None
    finally:
        sys.stderr = sys.__stderr__


_T = TypeVar("_T")
_P = ParamSpec("_P")


def wrap_async(
    pass_obj: bool = True, init_client: bool = True
) -> Callable[[Callable[_P, Awaitable[_T]]], Callable[_P, _T]]:
    def _decorator(callback: Callable[_P, Awaitable[_T]]) -> Callable[_P, _T]:
        assert iscoroutinefunction(callback)

        @functools.wraps(callback)
        def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> _T:
            with _runner() as runner:
                return runner.run(callback(*args, **kwargs))

        if pass_obj:
            wrapper = click.pass_obj(wrapper)  # type: ignore[assignment]

        return wrapper

    return _decorator


def option(*param_decls: Any, **attrs: Any) -> Callable[..., Any]:
    option_attrs = attrs.copy()
    typ = convert_type(attrs.get("type"), attrs.get("default"))
    autocompletion = getattr(typ, "complete", None)
    option_attrs.setdefault("shell_complete", autocompletion)
    return click.option(*param_decls, **option_attrs)


def argument(*param_decls: Any, **attrs: Any) -> Callable[..., Any]:
    arg_attrs = attrs.copy()
    typ = convert_type(attrs.get("type"), attrs.get("default"))
    autocompletion = getattr(typ, "complete", None)
    arg_attrs.setdefault("shell_complete", autocompletion)
    return click.argument(*param_decls, **arg_attrs)


BAKE_ID_PATTERN = r"bake-[0-9a-z]{8}-[0-9a-z]{4}-[0-9a-z]{4}-[0-9a-z]{4}-[0-9a-z]{12}"


async def resolve_bake(id_or_name: str, *, project: str, storage: Storage) -> str:
    if re.fullmatch(BAKE_ID_PATTERN, id_or_name):
        return id_or_name
    try:
        bake = await storage.project(yaml_id=project).bake(name=id_or_name).get()
        return bake.id
    except ResourceNotFound:
        return id_or_name



import apolo_sdk
import click
import signal
from apolo_cli.parse_utils import parse_timedelta
from contextlib import AsyncExitStack
from datetime import datetime, timezone
from dateutil.parser import isoparse
from typing import List, Optional, Sequence, Tuple

from apolo_flow.batch_runner import BatchRunner
from apolo_flow.cli.click_types import BAKE, BATCH, BATCH_OR_ALL, BakeTaskType
from apolo_flow.cli.utils import argument, option, resolve_bake, wrap_async
from apolo_flow.storage.api import ApiStorage
from apolo_flow.types import LocalPath

from ..parser import parse_bake_meta
from ..storage.base import Storage
from .root import Root


@click.command()
@option("--local-executor", is_flag=True, default=False, help="Run primary job locally")
@click.option(
    "-p",
    "--param",
    type=(str, str),
    multiple=True,
    help="Set params of the batch config",
)
@click.option(
    "-n",
    "--name",
    metavar="NAME",
    type=str,
    help="Optional bake name",
    default=None,
)
@click.option(
    "--meta-from-file",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
    help="File with params for batch.",
)
@click.option(
    "-t",
    "--tag",
    metavar="TAG",
    type=str,
    help="Optional bake tag, multiple values allowed",
    multiple=True,
)
@argument("batch", type=BATCH)
@wrap_async()
async def bake(
    root: Root,
    batch: str,
    local_executor: bool,
    meta_from_file: Optional[str],
    param: List[Tuple[str, str]],
    name: Optional[str],
    tag: Sequence[str],
) -> None:
    """Start a batch.

    Run BATCH pipeline remotely on the cluster.
    """
    async with AsyncExitStack() as stack:
        client = await stack.enter_async_context(apolo_sdk.get())
        storage: Storage = await stack.enter_async_context(ApiStorage(client))
        runner = await stack.enter_async_context(
            BatchRunner(root.config_dir, root.console, client, storage, root)
        )
        params = {key: value for key, value in param}
        if meta_from_file is not None:
            bake_meta = parse_bake_meta(LocalPath(meta_from_file))
            params = {**bake_meta, **params}
        await runner.bake(
            batch_name=batch,
            local_executor=local_executor,
            params=params,
            name=name,
            tags=tag,
        )


@click.command(hidden=True)
@click.argument("bake_id")
@wrap_async()
async def execute(
    root: Root,
    bake_id: str,
) -> None:
    """Start a batch.

    Run BATCH pipeline remotely on the cluster.
    """
    # apolo-flow execute is run in linux container only,
    # Linux signals are always defined.
    for signame in (
        signal.SIGHUP,
        signal.SIGINT,
        signal.SIGQUIT,
        signal.SIGTSTP,
        signal.SIGTERM,
        signal.SIGTTIN,
        signal.SIGTTOU,
        signal.SIGWINCH,
    ):
        # ignore everything, use apolo-flow cancel to stop the master job.
        signal.signal(signame, signal.SIG_IGN)
    async with AsyncExitStack() as stack:
        client = await stack.enter_async_context(apolo_sdk.get())
        storage: Storage = await stack.enter_async_context(ApiStorage(client))
        runner = await stack.enter_async_context(
            BatchRunner(root.config_dir, root.console, client, storage, root)
        )
        await runner.process(bake_id)


@click.command()
@click.option(
    "-t",
    "--tag",
    metavar="TAG",
    type=str,
    help="Filter out bakes by tag (multiple option)",
    multiple=True,
)
@option(
    "--since",
    metavar="DATE_OR_TIMEDELTA",
    help="Show bakes created after a specific date (including). "
    "Use value of format '1d2h3m4s' to specify moment in "
    "past relatively to current time.",
)
@option(
    "--until",
    metavar="DATE_OR_TIMEDELTA",
    help="Show bakes created before a specific date (including). "
    "Use value of format '1d2h3m4s' to specify moment in "
    "past relatively to current time.",
)
@option(
    "--recent-first/--recent-last",
    is_flag=True,
    default=False,
    help="Show newer bakes first or last",
)
@wrap_async()
async def bakes(
    root: Root,
    tag: Sequence[str],
    since: Optional[str],
    until: Optional[str],
    recent_first: bool,
) -> None:
    """List existing bakes."""
    async with AsyncExitStack() as stack:
        client = await stack.enter_async_context(apolo_sdk.get())
        storage: Storage = await stack.enter_async_context(ApiStorage(client))
        runner = await stack.enter_async_context(
            BatchRunner(root.config_dir, root.console, client, storage, root)
        )
        await runner.list_bakes(
            tags=set(tag),
            since=_parse_date(since),
            until=_parse_date(until),
            recent_first=recent_first,
        )


@click.command()
@argument("bake", type=BAKE)
@option(
    "-a",
    "--attempt",
    type=int,
    default=-1,
    help="Attempt number, the last attempt by default",
)
@click.option(
    "-o",
    "--output-graph",
    type=click.Path(file_okay=True, dir_okay=False, writable=True),
    help=(
        "A path to Graphviz (DOT) file. "
        "Autogenerated from BAKE_ID and attempt number by default"
    ),
)
@click.option(
    "--dot",
    is_flag=True,
    help=("Save DOT file with tasks statuses."),
)
@click.option(
    "--pdf",
    is_flag=True,
    help=("Save PDF file with tasks statuses."),
)
@click.option(
    "--view",
    is_flag=True,
    help=("Open generated PDF file with tasks statuses."),
)
@wrap_async()
async def inspect(
    root: Root,
    bake: str,
    attempt: int,
    output_graph: Optional[str],
    dot: bool,
    pdf: bool,
    view: bool,
) -> None:
    """Inspect a bake.

    Display a list of started/finished tasks of BAKE\\_ID.
    """
    async with AsyncExitStack() as stack:
        client = await stack.enter_async_context(apolo_sdk.get())
        storage: Storage = await stack.enter_async_context(ApiStorage(client))
        runner = await stack.enter_async_context(
            BatchRunner(root.config_dir, root.console, client, storage, root)
        )
        if output_graph is not None:
            real_output: Optional[LocalPath] = LocalPath(output_graph)
        else:
            real_output = None
        bake_id = await resolve_bake(bake, project=runner.project_id, storage=storage)

        await runner.inspect(
            bake_id,
            attempt_no=attempt,
            output=real_output,
            save_dot=dot,
            save_pdf=pdf,
            view_pdf=view,
        )


@click.command()
@argument("bake", type=BAKE)
@argument(
    "task_id",
    type=BakeTaskType(
        include_started=False,
        include_finished=True,
        bake_id_param_name="bake",
        attempt_no_param_name="attempt",
    ),
)
@click.option(
    "-a",
    "--attempt",
    type=int,
    default=-1,
    help="Attempt number, the last attempt by default",
)
@click.option(
    "-r/-R",
    "--raw/--no-raw",
    default=False,
    help=(
        "Raw mode disables the output postprocessing "
        "(the output is processed by default)"
    ),
)
@wrap_async()
async def show(
    root: Root,
    bake: str,
    attempt: int,
    task_id: str,
    raw: bool,
) -> None:
    """Show output of baked task.

    Display a logged output of TASK\\_ID from BAKE\\_ID.
    """
    async with AsyncExitStack() as stack:
        client = await stack.enter_async_context(apolo_sdk.get())
        storage: Storage = await stack.enter_async_context(ApiStorage(client))
        runner = await stack.enter_async_context(
            BatchRunner(root.config_dir, root.console, client, storage, root)
        )
        bake_id = await resolve_bake(bake, project=runner.project_id, storage=storage)
        await runner.logs(bake_id, task_id, attempt_no=attempt, raw=raw)


@click.command()
@argument("bake", type=BAKE)
@click.option(
    "-a",
    "--attempt",
    type=int,
    default=-1,
    help="Attempt number, the last attempt by default",
)
@wrap_async()
async def cancel(
    root: Root,
    bake: str,
    attempt: int,
) -> None:
    """Cancel a bake.

    Cancel a bake execution by stopping all started tasks.
    """
    async with AsyncExitStack() as stack:
        client = await stack.enter_async_context(apolo_sdk.get())
        storage: Storage = await stack.enter_async_context(ApiStorage(client))
        runner = await stack.enter_async_context(
            BatchRunner(root.config_dir, root.console, client, storage, root)
        )
        bake_id = await resolve_bake(bake, project=runner.project_id, storage=storage)
        await runner.cancel(bake_id, attempt_no=attempt)


@click.command()
@argument("batch", type=BATCH_OR_ALL)
@argument("task_id", type=str, required=False)
@wrap_async()
async def clear_cache(
    root: Root,
    batch: str,
    task_id: Optional[str],
) -> None:
    """Clear cache.

    Use `apolo-flow clear-cache <BATCH>` for cleaning up the cache for BATCH;
    Use `apolo-flow clear-cache <BATCH> <TASK_ID>` for cleaning up the cache
    for TASK_ID in BATCH;

    `apolo-flow clear-cache ALL` clears all caches.
    """
    async with AsyncExitStack() as stack:
        client = await stack.enter_async_context(apolo_sdk.get())
        storage: Storage = await stack.enter_async_context(ApiStorage(client))
        runner = await stack.enter_async_context(
            BatchRunner(root.config_dir, root.console, client, storage, root)
        )
        if batch == "ALL":
            await runner.clear_cache(None)
        else:
            await runner.clear_cache(batch, task_id)


@click.command()
@argument("bake", type=BAKE)
@option(
    "-a",
    "--attempt",
    type=int,
    default=-1,
    help="Attempt number, the last attempt by default",
)
@option("--local-executor", is_flag=True, default=False, help="Run primary job locally")
@option(
    "--from-failed/--no-from-failed",
    is_flag=True,
    default=True,
    help="Restart from the point of failure",
)
@wrap_async()
async def restart(
    root: Root,
    bake: str,
    attempt: int,
    from_failed: bool,
    local_executor: bool,
) -> None:
    """Start a batch.

    Run BATCH pipeline remotely on the cluster.
    """
    async with AsyncExitStack() as stack:
        client = await stack.enter_async_context(apolo_sdk.get())
        storage: Storage = await stack.enter_async_context(ApiStorage(client))
        runner = await stack.enter_async_context(
            BatchRunner(root.config_dir, root.console, client, storage, root)
        )
        bake_id = await resolve_bake(bake, project=runner.project_id, storage=storage)
        await runner.restart(
            bake_id,
            attempt_no=attempt,
            from_failed=from_failed,
            local_executor=local_executor,
        )


def _parse_date(value: Optional[str]) -> Optional[datetime]:
    if value:
        try:
            return isoparse(value)
        except ValueError:
            try:
                delta = parse_timedelta(value)
                return datetime.now(timezone.utc) - delta
            except click.UsageError:
                raise ValueError(
                    "Date should be either in ISO-8601 format or "
                    "relative delta of form 1d2h3m4s"
                )
    else:
        return None
